// Copyright 1996, 1997, 2000 by Jon Dart.  All Rights Reserved.

#include "stdafx.h"
#include "attacke.h"
#include "bearing.h"
#include "util.h"
#include "constant.h"
#include "debug.h"

static void addXRay(const Board &board, Bitmap & attacks,Square attack_square,
                    Square square,ColorType side)
{
   int dir = Bearing::Directions[attack_square][square];
   if (dir == 0)
      return;
   const int r = ((int)attack_square)/8;
   const int f = File(attack_square)-1;
   switch (dir)
   {
   case 1:
      // Find attacks on the rank
      {
         attacks.or(Bitmap::and(RankAttacksLeft(board,attack_square),
            board.rooksqueens[side]));
      }
      break;
   case -1:
      {
         attacks.or(Bitmap::and(RankAttacksRight(board,attack_square),
            board.rooksqueens[side]));
      }
      break;
   case 8:
      {
         attacks.or(Bitmap::and(FileAttacksUp(board,attack_square),board.rooksqueens[side]));
      }
      break;
   case -8:
      {
         attacks.or(Bitmap::and(FileAttacksDown(board,attack_square),board.rooksqueens[side]));
      }
      break;
   case 7:
      {
        Bitmap b(Bitmap::and(DiagAtcksA1(board,attack_square),
            DiagInfoA1[attack_square].lowerMask));
        b.and(board.bishopsqueens[side]);
        attacks.or(b);
      }
      break;
   case -7:
      {
         Bitmap b(Bitmap::and(DiagAtcksA1(board,attack_square),
             DiagInfoA1[attack_square].upperMask));
         b.and(board.bishopsqueens[side]);
         attacks.or(b);
      }
      break;
   case 9:
      {
         Bitmap b(Bitmap::and(DiagAtcksA8(board,attack_square),
            DiagInfoA8[attack_square].upperMask));
         b.and(board.bishopsqueens[side]);
         attacks.or(b);
      }
      break;
   case -9:
      {
         Bitmap b(Bitmap::and(DiagAtcksA8(board,attack_square),
            DiagInfoA8[attack_square].lowerMask));
         b.and(board.bishopsqueens[side]);
         attacks.or(b);
      }
      break;
   default:  
      break;
   }
}

int attack_estimate( const Board &board, const Move &emove )
{
   ASSERT(!IsNull(emove));
#ifdef ATTACK_TRACE
   cout << "attack_estimate " << MoveImage(emove) << endl;
#endif
   ColorType my_side = PieceColor(board[StartSquare(emove)]);
   ColorType side = my_side;
   Square square = (TypeOfMove(emove) == EnPassant) ?
                   board.EnPassantSq(OppositeColor(side)) : DestSquare(emove);
   Square attack_square = StartSquare(emove);
   Piece on_square = board[square];
   ASSERT(!IsEmptyPiece(on_square));
   Piece attacker = board[attack_square];
   ASSERT(!IsEmptyPiece(attacker));
   int score_count = 0;
   int score_list[20];
   int swap_score = 0;
   int gain;
   Bitmap opp_attacks(board.calc_attacks(square,OppositeColor(side)));
   if (opp_attacks.is_clear()) // piece is undefended
   {
      if (TypeOfMove(emove) == Promotion)
         return PieceValue(on_square) + PieceValues[PromoteTo(emove)] - 64;
      else
         return PieceValue(on_square);
   }
   Bitmap my_attacks(board.calc_attacks(square,side));
   my_attacks.clear(attack_square); // remove attacking piece
   if (TypeOfPiece(attacker) != Knight)
      addXRay(board,my_attacks,attack_square,square,side);
   int count = 0;
   for (;;)
   {
      attacker = board[attack_square];
#ifdef ATTACK_TRACE
      cout << " " << PieceImage(TypeOfPiece(attacker))
           << " on " <<
           FileImage(attack_square) <<
           RankImage(attack_square) <<
           " takes " << PieceImage(TypeOfPiece(on_square))
           << endl;
#endif
      gain = PieceValue(on_square);
      if (TypeOfPiece(attacker) == Pawn &&
          Rank(square,side) == 8)
      {
         if (count == 0) // initial capture is a promotion too
         {
            gain += (PieceValues[PromoteTo(emove)] - 64);
            on_square = MakePiece(PromoteTo(emove),side);
         }
         else
         {
            gain += (576-64); //Piece::Value(Piece::Queen) - Piece::Value(Piece::Pawn);
            on_square = MakePiece(Queen,side);
         }
      }
      else
         on_square = attacker;
      if (side == my_side)
         swap_score += gain;
      else
         swap_score -= gain;

      ASSERT(score_count < 20);
      score_list[score_count++] = swap_score;
      side = OppositeColor(side);
      if (side == my_side)
      {
         if (my_attacks.is_clear())
            break;
         attack_square = board.min_attacker(my_attacks,side).first_one();
         if (IsInvalid(attack_square))
            break;
         else
         {
            my_attacks.clear(attack_square);
            addXRay(board,my_attacks,attack_square,square,side);
         }
      }
      else
      {
         if (opp_attacks.is_clear())
            break;
         attack_square = board.min_attacker(opp_attacks,side).first_one();
         if (IsInvalid(attack_square))
            break;
         else
         {
            opp_attacks.clear(attack_square);
            addXRay(board,opp_attacks,attack_square,square,side);
         }
      }
      ++count;
   }
   if (score_count > 1 && score_count % 2 == 0)
   {
      // last capture is by other side, can't avoid it
      score_list[score_count-2] = score_list[score_count-1];
      --score_count; 
   }
   int i = score_count;
   while (i >= 3)
   {
      int max_score = Util::Max(score_list[i-1],score_list[i-2]);
      if (max_score < score_list[i-3])
         score_list[i-3] = max_score;
      i -= 2;
   }
#ifdef ATTACK__TRACE
   cout << "returning " << score_list[0] << endl;
#endif
   return score_list[0];
}

