// Copyright 1992, 1995 by Jon Dart.  All Rights Reserved.

#ifndef _BOOK_READER_H
#define _BOOK_READER_H

#include "bookdefs.h"
#include "bookinfo.h"
#include "board.h"
#include "move.h"
#include "pinfo.h"
#include "bookutil.h"

class Book_Reader
{
        // provides access to the opening book.

        public:

        Book_Reader(const char* fileName, const char* mappingName,
          bool create = false);
        // opens book file
                
        ~Book_Reader();
        // closes book file
                
        const BOOL is_open() const
        {
           return open;
        }
                
        const void head( const Board &b, enum styles style,
          Book_Location &loc);
        // Sets "loc" to the location of the head of the hash chain
        // for the board position.  If style is non-zero, selects moves
        // matching a particular style.

        void fetch( const Book_Location &loc, Book_Info &book_entry );
        // Fetches the book entry from the file. "loc" is the location.
        // Reads need not be sequential.
        
        void update(const Book_Location &loc, float learn_factor);
        // Updates a book position with a new weight.
        
        Move Book_Reader::pick( const Board &b, const Book_Location &loc,
          Book_Info &info);
        // Randomly pick a move for board position "b". "loc" is
        // the starting location, obtained from the head function.
        // "info" is filled in with information on the move chosen.
        
       int book_moves(const Board &b, Move *moves, const unsigned
       limit);
       
       protected:
               
       LPVOID pBook;
       byte *pPage;
       MEM_HANDLE hFileMapping;
       FILE_HANDLE book_file;
       BOOL open;
       int current_page;
       Book_Header hdr;
       
       private:
           
       void fetch_page(int page);
};

#endif
