// Copyright 1998 by Jon Dart. All Rights Reserved.

#include "stdafx.h"
#include "board.h"
#include "move.h"
#include "piece.h"
#include "movegen.h"

int legal_move(const Board &board, const Square &start,
               const Square &dest)
{
   // do a little basic sanity checking:
   if (!OnBoard(start) || !OnBoard(dest) || IsEmptyPiece(board[start]))
      return FALSE;

   PieceType promoteTo = InvalidPiece;
   if (TypeOfPiece(board[start]) == Pawn &&
       Rank(dest,board.Side()) == 8)
      promoteTo = Queen;
   Move emove = CreateMove(board,start,dest, promoteTo);
   Move moves[Move_Generator::MaxMoves];
   Move_Generator mg( board, 0, NullMove);
   int found = 0;
   int n = mg.Generate_Moves(moves, FALSE);

   for (int i = 0; i < n; i++)
   {
      if (MovesEqual(moves[i],emove))
      {
         found++;
         break;
      }
   }
   if (!found)
      return FALSE;
   else
   {
      // check for king en prise
      Board board_copy(board);
      const ColorType side = board.Side();
      board_copy.DoMove(emove);
      return !board_copy.any_attacks(
                                    board_copy.KingPos(side),OppositeColor(side));
   }
}

int have_legal_moves(const Board &board)
{
   Move moves[Move_Generator::MaxMoves];
   Move_Generator mg(board, 0, NullMove);
   int n = mg.Generate_Moves(moves, FALSE);

   for (int i = 0; i < n; i++)
   {
      // check for king en prise
      Board board_copy(board);
      const ColorType side = board.Side();
      board_copy.DoMove(moves[i]);
      if( !board_copy.any_attacks(
           board_copy.KingPos(side),OppositeColor(side)))
      {
         return 1;
      }
   }
   return 0;
}

