// Copyright 1992, 1997, 1999 by Jon Dart.  All Rights Reserved.
#ifndef _PIECE_H
#define _PIECE_H

#include "types.h"
#include "color.h"

typedef int Piece;

typedef int PieceType;

#define EmptyPiece 0
#define Pawn 1
#define Knight 2
#define Bishop 3
#define Rook 4
#define Queen 5
#define King 6
#define InvalidPiece 7

#define PAWN_VALUE 64
#define BISHOP_VALUE 200
#define KING_VALUE 2000

extern const unsigned PieceValues[8];
extern const int _sliders[16];

inline BOOL sliding(Piece p) 
{
   return _sliders[p];
}

#define MakePiece( type, color ) _pieces[type][color]
#define MakeWhitePiece( type ) (((Piece)type) + 8)
#define MakeBlackPiece( type ) ((Piece)type)
    
#define TypeOfPiece(piece) ((PieceType)(piece & 7))
   
#define PieceColor(piece) ((piece & 8) ? White : Black)
#define ColorOfPiece(piece) ((piece & 8) ? White : Black)

#define IsEmptyPiece(piece) (TypeOfPiece(piece) == EmptyPiece)
   
#define PieceValue(piece) PieceValues[TypeOfPiece(piece)]   

extern PieceType PieceCharValue( const char );

// 1-character representation of piece
extern char PieceImage(const PieceType p);

extern const Piece _pieces[8][2];	   

#define whitePawn 9
#define whiteKnight 10
#define whiteBishop 11
#define whiteRook 12
#define whiteQueen 13
#define whiteKing 14
#define blackPawn 1
#define blackKnight 2
#define blackBishop 3
#define blackRook 4
#define blackQueen 5
#define blackKing 6
#endif


