// Copyright 1999 by Jon Dart.  All Rights Reserved.

#include "posnbook.h"
#include "globals.h"
#include "util.h"

#define NUM_ENTRIES 32768

PositionBook::PositionBook()
: book_file(NULL),
pBook(NULL),
pPage(NULL),
hFileMapping(NULL),
open(FALSE)
{                 
   const char *fileName = POSITION_BOOK_NAME;
   const char *mappingMap = POSITION_MAPPING_NAME;
   static char book_path[MAX_PATH];
   const int learn_size = NUM_ENTRIES*sizeof(Position_Info)+sizeof(Book_Header);
   int newBook = 0;
   strcpy(book_path,programPath);
   {
      // find the last part of the program pathname
      char *p = strrchr(book_path,'\\');
      if (p)
      {
         // replace it with the book file name
         strcpy(p+1,fileName);
      }
      else
         strcpy(book_path,fileName);
      book_file = BookUtil::openFile(book_path,false);
      if (book_file == NULL)
      {
         // create new learning file
         if (theLog) theLog->write("creating position.bin\n");
         book_file = BookUtil::createFile(book_path,learn_size
           );
         ++newBook;
      }
   }
   if (book_file != NULL)
   {
      hFileMapping = BookUtil::createMap(book_file,false);
   }
   if (hFileMapping)
   {
      pBook = BookUtil::map(hFileMapping,0,learn_size,false);
      if (newBook)
      {
         *((char*)pBook) = Book_Version;
      }
   }

   open = pBook != NULL;
   if (!open)
      return;

   char *p = (char*)pBook;

   if (*p != (byte)Book_Version)
   {
      char msg[128];
      sprintf(msg,"Expected book version %d, got %d instead",
              (int)Book_Version,(int)*p);
      MessageBox(NULL,msg,"Warning",MB_ICONEXCLAMATION);
      BookUtil::unmap(pBook);
      BookUtil::closeMap(hFileMapping);
      BookUtil::closeFile(book_file);
      open = FALSE;
      pBook = NULL;
      return;
   }
}

PositionBook::~PositionBook()
{
   if (pBook)
      BookUtil::unmap(pBook);
   if (hFileMapping)
      BookUtil::closeMap(hFileMapping);
   if (book_file)
      BookUtil::closeFile(book_file);
}

void PositionBook::add(const Position_Info &info)
{
   hash_t index = info.hash_code() % NUM_ENTRIES;
   char *p = (char*)pBook + sizeof(Book_Header) + index*sizeof(Position_Info);
   memcpy(p,&info,sizeof(Position_Info));
}

int PositionBook::size() const
{
   return NUM_ENTRIES; // currently fixed
}
        
Position_Info * PositionBook::lookup(const Board &board,int rep_count)
{
   Position_Info look(board,rep_count);
   hash_t index = quickmod(look.hash_code(),NUM_ENTRIES);
   Position_Info *p = 
    (Position_Info*)((char*)pBook + sizeof(Book_Header) + index*sizeof(Position_Info));
   if (p->hash_code() == look.hash_code())
   {
     return p;
   }
   else
     return NULL;
}
