// Copyright 1994-1999 by Jon Dart.  All Rights Reserved.

#include "stdafx.h"
#include "scoring.h"
#include "bearing.h"
#include "constant.h"
#include "util.h"
#include "pawnentr.h"
#include "kingpawn.h"
#include "search.h"
#include "bhash.h"
#include "config.h"

//#define EVAL_DEBUG
//#define PAWN_DEBUG
//#define LAZY_DEBUG

static const int KnightCenterScores[] =
{-4, 0, 0, 0, 0, 0, 0, -4,
   -4, 0, 2, 2, 2, 2, 0, -4,
   -4, 2, 3, 2, 2, 3, 2, -4,
   -4, 2, 2, 5, 5, 2, 2, -4,
   -4, 2, 2, 5, 5, 2, 2, -4,
   -4, 2, 3, 2, 2, 3, 2, -4,
   -4, 0, 2, 2, 2, 2, 0, -4,
   -4, 0, 0, 0, 0, 0, 0, -4};

static const int BishopCenterScores[] =
{-2, -2, -2, -2, -2, -2, -2, -2,
   -2, 0, 0, 0, 0, 0, 0, -2,
   -2, 0, 1, 1, 1, 1, 0, -2,
   -2, 0, 1, 2, 2, 1, 0, -2,
   -2, 0, 1, 2, 2, 1, 0, -2,
   -2, 0, 1, 1, 1, 1, 0, -2,
   -2, 0, 0, 0, 0, 0, 0, -2,
   -2, -2, -2, -2, -2, -2, -2, -2};

static const int BishopPositionB[] =
{-3, -6, -6, -6, -6, -6, -6, -3,  
   0, 3, 0, 0, 0, 0, 3, 0,
   0, 0, 3, 0, 0, 3, 0, 0,
   0, 0, 0, 3, 3, 0, 0, 0,
   0, 0, 0, 3, 3, 0, 0, 0,
   0, 0, 3, 0, 0, 3, 0, 0,
   0, 3, 0, 0, 0, 0, 3, 0,
   3, 0, 0, 0, 0, 0, 0, 3};

static const int BishopPositionW[] =
{3, 0, 0, 0, 0, 0, 0, 3,  
   0, 3, 0, 0, 0, 0, 3, 0,
   0, 0, 3, 0, 0, 3, 0, 0,
   0, 0, 0, 3, 3, 0, 0, 0,
   0, 0, 0, 3, 3, 0, 0, 0,
   0, 0, 3, 0, 0, 3, 0, 0,
   0, 3, 0, 0, 0, 0, 3, 0,
   -3, -6, -6, -6, -6, -6, -6, -3};


static const int PawnCenterScores[] =
{0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 4, 4, 0, 0, 0,
   0, 0, 5, 7, 7, 5, 0, 0,
   0, 0, 3, 4, 4, 3, 0, 0,
   0, 0, -1, -2, -2, -1, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0};

static const int KingCenterScores[] =
{0, 0, 1, 1, 1, 1, 0, 0,
   0, 1, 2, 3, 3, 2, 1, 0,
   1, 2, 6, 7, 7, 6, 2, 1,
   1, 3, 7, 9, 9, 7, 3, 1,
   1, 3, 7, 9, 9, 7, 3, 1,
   1, 2, 6, 7, 7, 6, 2, 1,
   0, 1, 2, 3, 3, 2, 1, 0,
   0, 0, 1, 1, 1, 1, 0, 0};

static const int CenterPawnCoverInit[] =
{0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 1, 1, 1, 1, 0, 0,
   0, 0, 1, 1, 1, 1, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0};

static const int QueenCenterScores[] =
{-3, 0, 0, 0, 0, 0, 0, -3,
   -2, 0, 0, 0, 0, 0, 0, -2,
   -2, 0, 1, 1, 1, 1, 0, -2,
   -2, 0, 1, 2, 2, 1, 0, -2,
   -2, 0, 1, 2, 2, 1, 0, -2,
   -2, 0, 1, 1, 1, 1, 0, -2,
   -2, 0, 0, 0, 0, 0, 0, -2,
   -3, 0, 0, 0, 0, 0, 0, -3};

static const int OutpostScoresB[] =
{0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 2, 2, 0, 0, 0,
0, 0, 2, 5, 5, 2, 0, 0,
0, 0, 3, 6, 6, 3, 0, 0,
0, 0, 1, 2, 2, 1, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0 };

static const int OutpostScoresW[] =
{0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 1, 2, 2, 1, 0, 0,
0, 0, 3, 6, 6, 3, 0, 0,
0, 0, 2, 5, 5, 2, 0, 0,
0, 0, 0, 2, 2, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0 };

static const int KBNKScores[] =
{-20, -15, -10, -3, 3, 10, 15, 20,
   -15, -10, -3, 0, -1, 3, 10, 15,
   -10, -3, 0, 0, 0, -1, 3, 10,
   -3, -3, 0, -2, -2, 0, -1, 3,
   3, -1, 0, -2, -2, 0, 0, -3,
   10, 3, -1, 0, 0, 0, -3, -10,
   15, 10, 3, -1, 0, -3, -10, -15,
   20, 15, 10, 3, -3, -10, -15, -20};

static const int quadrants[] =
{ 0, 0, 0, 0, 1, 1, 1, 1,
   0, 0, 0, 0, 1, 1, 1, 1,
   0, 0, 0, 0, 1, 1, 1, 1,
   0, 0, 0, 0, 1, 1, 1, 1,
   2, 2, 2, 2, 3, 3, 3, 3,
   2, 2, 2, 2, 3, 3, 3, 3,
   2, 2, 2, 2, 3, 3, 3, 3,
   2, 2, 2, 2, 3, 3, 3, 3};

static BOOL endgame;
static ColorType side_to_move;

// Weights for positional scoring components.  A plus value means that
// the item being scored is valuable to the player; a minus value means
// that a penalty is extracted.  In most cases, scoring components are
// simply summed together to get a total score, but some are folded
// in, in a more complex way.

// material terms
static const int TWO_BISHOPS = 20;
static const int BAD_TRADE=-50;
// development terms
static const int BLOCKED_BISHOP = -4;
static const int BISHOP_MOBILITY[] = {-8, -8, -6, -3, 0, 2, 4, 6, 8, 8, 8 };
static const int BAD_BISHOP = -5;
static const int CENTER_PAWN_BLOCK = -6;  // e or d pawn blocked by piece
static const int LONG_DIAG = 3; // bishop on long diagonal
static const int KNIGHT_BACK = -7;
static const int ROOK_TRAPPED = -15; // rook is trapped by king
static const int KNIGHT_OUTPOST = 2; // in addition to reg. outpost scorea
static const int KNIGHT_MOBILITY[] = {-6, -1, 1, 2, 3, 4, 5, 6, 7};
static const int ROOK_ON_7TH_RANK = 15;
static const int QUEEN_ON_7TH_RANK = 6;
static const int TWO_ROOKS_ON_7TH_RANK = 15;
static const int QUEEN_DEVELOP[] = {6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3,
   2, 2};
static const int QUEEN_OUT = -7; // premature development of queen
static const int DOUBLED_ROOKS = 8; // on open file
static const int ROOK_ON_OPEN_FILE = 6;
static const int QUEEN_ON_OPEN_FILE = 4;
// castling terms
static const int KCASTLE = 11;
static const int QCASTLE = 8;
static const int CAN_CASTLE = -2;
static const int CANT_CASTLE = -12;
static const int CANT_CASTLE_K_SIDE = -6;
static const int CANT_CASTLE_Q_SIDE = -4;
// pawn structure terms
static const int WEAK = -4;
static const int WEAK2 = -4;
static const int WEAK_ON_OPEN_FILE = -8;
static const int SPACE=2;
// isolated pawn - note: these penalties are low because an
// isolated pawn is also "weak".
static const int ISOLATED_PAWN[] = { 0, -3, -7, -12, -18, -24, -30, -36 };
static const int DOUBLED_PAWNS = -12;
static const int TRIPLED_PAWNS = -25;
static const int PASSED_PAWN[8] = { 0, 10, 13, 17, 24, 38, 65, 0}; // score by rank
static const int PASSED_PAWN_BLOCK[8] =
{ 2, 2, 4, 5, 6, 7, 7, 0};
// outside passed pawns 
static const int OUTSIDE_PP[] = { 35, 28, 23, 17, 13, 10,
   7, 5, 4, 3, 3, 3, 3};
static const int SUPPORTED_PASSER6 = 24;
static const int SUPPORTED_PASSER7 = 48;
static const int PASSER_MATERIAL_BONUS[] =
 { 12, 8, 5, 3, 2, 1, 0, -2, -4, -6, -6, -6, -6, -6 };
static const int PASSED_PAWN_CANT_ADVANCE = 4;
static const int ROOK_BEHIND_PP[8] = 
{ 0, 2, 4, 6, 8, 9, 10, 0};
static const int CONNECTED_PASSERS[8] = 
//{0, 6, 7, 8, 10, 16, 24, 0};
{0, 6, 7, 8, 20, 37, 75, 0};
static const int ADV_CONNECTED_PASSERS = 25;
static const int RIM_PASSED_PAWN = -7; // pawn on h-file or a-file
static const int PAWN_DUO = 5; // a la Kmoch
static const int ADV_UNSUPPORTED_PAWN = -4; // an advanced but unsupported pawn
static const int STONEWALL = -8; // opponent has stonewall formation

// king safety terms
static const int ROOK_FILE_MATE_THREAT = -216; // mating threat on rook file
static const int POTENTIAL_ROOK_FILE_MATE = -50; // mating threat on rook file
static const int LONG_DIAGONAL_THREAT = -4;
static const int DIAGONAL_MATE_THREAT = 0; //-50;
// king off back rank penalty, increases with opponent's material
static const int KING_OFF_BACK_RANK[] = {0 , 0, -1, -3, -4, -5, -6, -7,
   -9, -9, -10, -11, -12};
static const int KS_MATERIAL_SCALE[] =
  { 0, 0, 1, 4, 7, 10, 
   12, 13, 14, 15, 16, 16, 16 };
// king pawn cover penalty, increases with opponent's material
static const int FIANCHETTO = 6;
static const int BP_COVER_MISSING = -9;
static const int NP_COVER_MISSING = -11;
static const int RP_COVER_MISSING = -11;
static const int BP_COVER_MISSING2 = -23;
static const int NP_COVER_MISSING2 = -28;
static const int RP_COVER_MISSING2 = -28;
static const int KING_NEAR_OPEN_FILE = -8;
static const int KING_QUEEN_DISTANCE[] = {8, 6, 4, 3, 2, 1, 0, -1};

static const int BATTERY0 = 0; //-2
static const int BATTERY1 = -4;
static const int BATTERY2 = -12;
static const int BATTERY3 = -20;
static const int FILE_DANGER1 = -3; // rook on file
static const int FILE_DANGER2 = -5; // queen on file
static const int FILE_DANGER3 = -7; // rook on open file
static const int FILE_DANGER4 = -12; // queen on open file
static const int KING_TRAPPED_IN_CORNER = -5; 
static const int KING_PIECE_PROXIMITY[10] = 
{ 0, -4, -14, -30, -48, 
   -48, -48, -48, -48, -48};
// king position terms - valid only in endgame
static const int ADVANCED_KING = 4;
static const int OPPOSITION = 6; // twice this for direct opposition
// piece safety terms
static const int BISHOP_TRAPPED = -180; // bishop trapped by pawn
// general endgame terms
static const int KING_PAWN_DISTANCE = -10; // distance from opposing king
// KP vs. K endgame
static const int UNCATCHABLE_PAWN = 250;
// a pawn that may become "uncatchable"
static const int UNCATCHABLE_PAWN2[] = {24, 24, 18, 12, 8, 4, 0, 0, 0, 0,
0, 0, 0};
static const int KING_AHEAD_OF_PAWN = 32;
static const int DISTANCE_FROM_PAWN = -8;
// Bishop vs. Knight with pawns on both sides
static const int BISHOP_ENDGAME_BONUS = 10;

static Bitmap passedW[64];
static Bitmap passedB[64];
static Bitmap backwardW[64];
static Bitmap backwardB[64];
static Bitmap weakW[64];
static Bitmap weakB[64];
static Bitmap outpostW[64];
static Bitmap outpostB[64];
static Bitmap right_side_mask[8], left_side_mask[8];
// area near king 
static Bitmap king_attacks[4];
static Bitmap king_near_proximity[64];
static Bitmap w_pawn_attacks[8], b_pawn_attacks[8];
static Bitmap center_squares;
static Bitmap white_squares;
static Bitmap black_squares;
static Bitmap center_pawn_cover_b;
static Bitmap center_pawn_cover_w;
static Bitmap rook_pawn_mask;
static Bitmap longdiagDa1,longdiagDa8;
static Bitmap w_bishop_trap, b_bishop_trap;
static Bitmap w_stonewall, b_stonewall, white_rook_trapped,
black_rook_trapped;
static Bitmap right_mask,left_mask;
static Bitmap w_cover_mask[64], b_cover_mask[64],
              w_cover_mask2[64], b_cover_mask2[64];
static int done_init = 0;
static Pawn_Entry *w_pawn_hash_table = NULL;
static Pawn_Entry *b_pawn_hash_table = NULL;
static King_Pawn_Entry *king_pawn_hash_table = NULL;
static Bitmap ghfiles;
static int opposition[64][64];
int distances[64][64];

static int material_level[2];
static Pawn_Entry *hash_entry_w, *hash_entry_b;
static byte first_passer_w, last_passer_w,
first_passer_b, last_passer_b, b_protected_pp, w_protected_pp;
// most advanced passer that is "uncatchable". 
static int white_passer_rank, black_passer_rank;
// most advanced passer that is "uncatchable" (do not count
// separated pawns that are not individually uncatchable)
static int white_passer_race_rank, black_passer_race_rank;
static Square white_passer_sq, black_passer_sq;
static Square white_passer_race_sq, black_passer_race_sq;

static inline int distance(int x, int y)
{
   return distances[x][y];
}

#define PAWN_HASH_SIZE 12500
#define KING_PAWN_HASH_SIZE 15000

static void init_bitmaps()
{
   int i, r,f;
   for (i = 0; i < 64; i++)
   {
      for (int j = 0; j < 64; j++)
      {
         int file_distance, rank_distance;
         Square kp = i;
         Square oppkp = j;
         file_distance=Util::Abs(File(kp)-File(oppkp));
         rank_distance=Util::Abs(Rank(kp,Black)-Rank(oppkp,Black));
         if (rank_distance == 0 && file_distance == 2)
         {
            opposition[i][j] = 2; // direct opposition
         } else if (file_distance == 0 && rank_distance % 2 == 0)
            opposition[i][j] = 1; // indirect opposition
         else if ((file_distance == 2) && (rank_distance == 2))
            opposition[i][j] = 1;
         else
            opposition[i][j] = 0;
         if (file_distance > rank_distance)
            distances[i][j] = file_distance;
         else
            distances[i][j] = rank_distance;
         if (distances[i][j] <= 2)
         {
            king_near_proximity[i].set(j);
         }
      }
   }
   for (Square sq = 0; sq < 64; sq++)
   {
      if (InCenter(sq))
      {
         center_squares.set(sq);
      }
      if (SquareColor(sq) == White)
         white_squares.set(sq);
      else
         black_squares.set(sq);
      if (CenterPawnCoverInit[sq])
      {
         center_pawn_cover_w.set(sq);
         center_pawn_cover_b.set(63-sq);
      }
      int rank = Rank(sq,White);
      int rankb= Rank(sq,Black);
      int file = File(sq);
      if (file == 1 || file == 8)
         rook_pawn_mask.set(sq);
      for (r = rank; r > 1; r--)
      {
         Square sq2 = Square(file,r,White);
         if (file != 8)
         {
            sq2 = Square(file+1,r,White);
            backwardW[sq].set((int)sq2);
         }
         if (file != 1)
         {
            sq2 = Square(file-1,r,White);
            backwardW[sq].set((int)sq2);
         }
      }
      weakW[sq] = backwardW[sq];
      if (file != 8)
      {
         weakW[sq].clear(Square(file+1,rank,White));
      }
      if (Rank(sq,White) < 3)
      {
         w_cover_mask[sq].set(Square(file,rank+1,White));
         w_cover_mask2[sq].set(Square(file,rank+2,White));
         if (file != 1)
         {
            w_cover_mask[sq].set(Square(file-1,rank+1,White));
            w_cover_mask2[sq].set(Square(file-1,rank+2,White));
            w_cover_mask2[sq].set(Square(file-1,rank,White));
         }
         if (file != 8)
         {
            w_cover_mask[sq].set(Square(file+1,rank+1,White));
            w_cover_mask2[sq].set(Square(file+1,rank+2,White));
            w_cover_mask2[sq].set(Square(file+1,rank,White));
         }
      }
      if (rankb < 3)
      {
         b_cover_mask[sq].set(Square(file,rankb+1,Black));
         b_cover_mask2[sq].set(Square(file,rankb+2,Black));
         if (file != 1)
         {
            b_cover_mask[sq].set(Square(file-1,rankb+1,Black));
            b_cover_mask2[sq].set(Square(file-1,rankb+2,Black));
            b_cover_mask2[sq].set(Square(file-1,rankb,Black));
         }
         if (file != 8)
         {
            b_cover_mask[sq].set(Square(file+1,rankb+1,Black));
            b_cover_mask2[sq].set(Square(file+1,rankb+2,Black));
            b_cover_mask2[sq].set(Square(file+1,rankb,Black));
         }
      }
      if (file != 1)
         weakW[sq].clear(Square(file-1,rank,White));
      for (r = rank + 1; r < 8; r++)
      {
         Square sq2 = Square(file,r,White);
         passedW[sq].set(sq2);
         if (file != 8)
         {
            sq2 = Square(file+1,r,White);
            passedW[sq].set(sq2);
            outpostW[sq].set(sq2);
         }
         if (file != 1)
         {
            sq2 = Square(file-1,r,White);
            passedW[sq].set(sq2);
            outpostW[sq].set(sq2);
         }
      }
      rank = Rank(sq,Black);
      for (r = rank; r > 1; r--)
      {
         Square sq2 = Square(file,r,Black);
         if (file != 8)
         {
            sq2 = Square(file+1,r,Black);
            backwardB[sq].set(sq2);
         }
         if (file != 1)
         {
            sq2 = Square(file-1,r,Black);
            backwardB[sq].set(sq2);
         }
      }
      weakB[sq] = backwardB[sq];
      if (file != 8)
         weakB[sq].clear(Square(file+1,rank,Black));
      if (file != 1)
         weakB[sq].clear(Square(file-1,rank,Black));
      for (r = rank + 1; r < 8; r++)
      {
         Square sq2 = Square(file,r,Black);
         passedB[sq].set((int)sq2);
         if (file != 8)
         {
            sq2 = Square(file+1,r,Black);
            passedB[sq].set(sq2);
            outpostB[sq].set(sq2);
         }
         if (file != 1)
         {
            sq2 = Square(file-1,r,Black);
            passedB[sq].set(sq2);
            outpostB[sq].set(sq2);
         }
      }
   }
   for (r = 1; r <= 4; r++)
   {
      for (f = 1; f<=4; f++)
      {
         Square sq = Square(f,r,Black);
         king_attacks[0].set(sq);
      }
   }
   for (r = 1; r <= 4; r++)
   {
      for (f = 5; f<=8; f++)
      {
         Square sq = Square(f,r,Black);
         king_attacks[1].set(sq);
      }
   }
   for (r = 1; r <= 4; r++)
   {
      for (f = 1; f<=4; f++)
      {
         Square sq = Square(f,r,White);
         king_attacks[2].set(sq);
      }
   }
   for (r = 1; r <= 4; r++)
   {
      for (f = 5; f<=8; f++)
      {
         Square sq = Square(f,r,White);
         king_attacks[3].set(sq);
      }
   }
   for (int file = 1; file < 9; file++)
   {
      if (file < 4)
      {
         for (int r = 2; r <= 5; r++)
         {
            w_pawn_attacks[file-1].set((int)Square(1,r,White));
            w_pawn_attacks[file-1].set((int)Square(2,r,White));
            w_pawn_attacks[file-1].set((int)Square(3,r,White));
            b_pawn_attacks[file-1].set((int)Square(1,r,Black));
            b_pawn_attacks[file-1].set((int)Square(2,r,Black));
            b_pawn_attacks[file-1].set((int)Square(3,r,Black));
         }
      } 
      else if (file > 5)
      {
         for (int r = 2; r <= 5; r++)
         {
            w_pawn_attacks[file-1].set((int)Square(6,r,White));
            w_pawn_attacks[file-1].set((int)Square(7,r,White));
            w_pawn_attacks[file-1].set((int)Square(8,r,White));
            b_pawn_attacks[file-1].set((int)Square(6,r,Black));
            b_pawn_attacks[file-1].set((int)Square(7,r,Black));
            b_pawn_attacks[file-1].set((int)Square(8,r,Black));
         }
      } 
      else
      {
         for (int r = 2; r <= 5; r++)
         {
            w_pawn_attacks[file-1].set((int)Square(4,r,White));
            w_pawn_attacks[file-1].set((int)Square(5,r,White));
            b_pawn_attacks[file-1].set((int)Square(4,r,Black));
            b_pawn_attacks[file-1].set((int)Square(5,r,Black));
         }       
      }
   }
   longdiagDa1= Bitmap(0);
   for (int x = 56; x >= 0; x-=7)
      longdiagDa1.set(x);
   longdiagDa8= Bitmap(0);
   for (x = 0; x <= 63; x+=9)
      longdiagDa8.set(x);
   for (x=1; x <8; x++)
   {
      for (int y=x-1;y>=0;y--)
         left_side_mask[x].or(Bearing::file_mask[y]);
   }
   for (x=6; x >=0; x--)
   {
      for (int y=x+1;y<8;y++)
         right_side_mask[x].or(Bearing::file_mask[y]);
   }
   b_bishop_trap.set(55);
   b_bishop_trap.set(62);
   b_bishop_trap.set(48);
   b_bishop_trap.set(57);
   w_bishop_trap.set(1);
   w_bishop_trap.set(8);
   w_bishop_trap.set(6);
   w_bishop_trap.set(15);
   w_stonewall.set(35);
   w_stonewall.set(37);
   b_stonewall.set(27);
   b_stonewall.set(29);
   black_rook_trapped.set(7);
   black_rook_trapped.set(15);
   black_rook_trapped.set(6);
   white_rook_trapped.set(63);
   white_rook_trapped.set(63-8);
   white_rook_trapped.set(62);
   ghfiles = Bitmap::or(Bearing::file_mask[7],Bearing::file_mask[6]);
   left_mask = ghfiles;
   left_mask.or(Bearing::file_mask[5]);
   right_mask = Bitmap::or(Bearing::file_mask[0],Bearing::file_mask[1]);
   right_mask.or(Bearing::file_mask[2]);
}

// N.b. file is 0..7
inline int PFileCount( const Board &board, const int file, const ColorType side ) 
{
   return Bitmap::and(board.pawn_bits[side],Bearing::file_mask[file]).
     bit_count();
}

// N.b. file is 0..7
inline int FileOpen( const int file, const ColorType side ) 
{
   return ((side == White) ? hash_entry_w->open_files :
                            hash_entry_b->open_files) & (1<<file);
}

BOOL uncatchable(const Board &board,
                 const Square pawnpos,
                 const Square kp,
                 const Square oppkp,
                 const ColorType side)
{
   // Use the "square of the king" rule to see whether a passed pawn can
   // be captured before it promotes (or immediately after).

   int oppkp_rank = Rank(oppkp,side);
   int rankdist = Rank(pawnpos,side) - oppkp_rank;

   if (board.Side() == side)
   {
      if (rankdist > 0)
      {
         // If the king is behind us and we are on move we can't be caught.
         return TRUE;
      } 
      else
      {
         if (
            8 - Rank(pawnpos,side) < distance(oppkp,Square(File(pawnpos),8,side)))
            return TRUE;
      }
   } 
   else
   {
      if (rankdist > 1)
      {
         // pawn > 1 square ahead
         return TRUE;
      } 
      else
      {
         if (
            9 - Rank(pawnpos,side) < distance(oppkp,Square(File(pawnpos),8,side)))
            return TRUE;
      }
   }
   // The opposing king could theoretically catch the pawn.  Now
   // see if our king is in a position to prevent this.
   int kpfile = File(kp);
   int pawnfile = File(pawnpos);
   int kprank = Rank(kp,side);
   int pawnrank = Rank(pawnpos,side);
   if (kprank < pawnrank)
   {
      // our king is behind our pawn
      return FALSE;
   }
   int d;
   if (pawnfile == 1)
   {
      // rook pawn.
      if (kpfile == 1 && File(oppkp) == 3)
      {
         return (board.Side() == side && pawnrank < 7 &&
             kprank > oppkp_rank &&
            (kprank-pawnrank <= 3));
      }
      if (File(oppkp) == 1 || 
          distance(oppkp,pawnpos) - distance(kp,pawnpos) < 0)
      {
         // opposing king is ahead of the pawn, or closer
         // than our own king.
         return FALSE;
      }
      if (side == White)
      {
         if (oppkp <= 2)
            return FALSE;
         d = distance(oppkp,2) - distance(kp,9);
         if (board.Side() == side && d >= 0)
         {
             return TRUE;
         }
         if (board.Side() != side && d > 0)
         {
            return TRUE;
         }
      } 
      else
      {
         if (oppkp >= 56 && oppkp <= 58)
            return FALSE;
         d = distance(oppkp,58) - distance(kp,49);
         if (board.Side() == side && d >= 0)
         {
             return TRUE;
         }
         if (board.Side() != side && d > 0)
         {
            return TRUE;
         }
      }
   } 
   else if (pawnfile == 8)
   {
      if (kpfile == 8 && File(oppkp) == 6)
      {
         return (board.Side() == side && pawnrank < 7 &&
             kprank > oppkp_rank &&
            (kprank-pawnrank <= 3));
      }
      if (File(oppkp) == 8 || 
          distance(oppkp,pawnpos) - distance(kp,pawnpos) < 0)
      {
         // opposing king is ahead of the pawn, or closer
         // than our own king.
         return FALSE;
      }
      if (side == White)
      {
         if (oppkp >= 5 && oppkp <= 7)
            return FALSE;
         d = distance(oppkp,5) - distance(kp,14);
         if (board.Side() == side && d >= 0)
         {
             return TRUE;
         }
         if (board.Side() != side && d > 0)
         {
            return TRUE;
         }
      } 
      else
      {
         if (oppkp >= 61)
            return FALSE;
         d = distance(oppkp,61) - distance(kp,54);
         if (board.Side() == side && d >= 0)
         {
             return TRUE;
         }
         if (board.Side() != side && d > 0)
         {
            return TRUE;
         }
      }
   } 
   else
   {
      // not a rook pawn
         // If our king is >=2 squares ahead and the opposing king
         // is not closer, it is a win
         if (distance(kp,pawnpos) <= distance(oppkp,pawnpos))
         {
            if (kprank > pawnrank + 1)
               return TRUE;
         }
         // a supported pawn on the 6th rank or past is uncatchable
         if (pawnrank >=6  && kprank >= 6)
         {
            Bitmap ka(Bearing::king_attacks[pawnpos]);
            ka.and(board.king_bits[side]);
            if (!ka.is_clear())
               return TRUE;
         }
         // If the king is directly ahead of the pawn and we
         // have the opposition, it is a win
         if (board.Side() != side && kp == pawnpos+Direction[side]*8)
         {
            if (File(oppkp) == pawnfile &&
                ((int)oppkp - (int)kp) % 16 == 0)
               return TRUE;
         }
         // If we guard the pawn and the opposing king is behind or
         // on the same rank, we win
         if (rankdist >= 0 && (kpfile == pawnfile-1 ||
                               kpfile == pawnfile+1) && (kprank == pawnrank ||
                                                         kprank == pawnrank+1))
            return TRUE;
   }
   return FALSE;
}   

// evaluate outside passed pawns.
static int eval_outside_pp(const Board &board)
{
   // This code is based on similar code in Crafty.  See also
   // Fine's Basic Chess Endings, pp. 35-36.
   int score = 0;
   Bitmap all_pawns(Bitmap::or(board.pawn_bits[White],board.pawn_bits[Black]));
   int b_outside = 0, w_outside = 0;
   if (first_passer_w != 9)
   {
      if ((first_passer_w < first_passer_b-1) ||
          (last_passer_w > last_passer_b+1))
      {
         if (first_passer_w <= 4)
         {
            if (!Bitmap::and(all_pawns,right_side_mask[first_passer_w-1]).is_clear()
                &&  Bitmap::and(board.pawn_bits[Black],left_side_mask[first_passer_w-1]).is_clear())
            {
               if (!b_protected_pp || w_protected_pp)
               {
                  w_outside++;
                  goto step2;
               }
            }
         }
         if (last_passer_w >= 5)
         {
            if (!Bitmap::and(all_pawns,left_side_mask[last_passer_w-1]).is_clear()
                &&  Bitmap::and(board.pawn_bits[Black],right_side_mask[last_passer_w-1]).is_clear())
            {
               if (!b_protected_pp || w_protected_pp)
               {
                  w_outside++;
               }
            }
         }
      }
   }
   step2:
   if (first_passer_b != 9)
   {
      if ((first_passer_b < first_passer_w-1) ||
          (last_passer_b > last_passer_w+1))
      {
         if (first_passer_b <= 4)
         {
            if (!Bitmap::and(all_pawns,right_side_mask[first_passer_b-1]).is_clear()
                &&  Bitmap::and(board.pawn_bits[White],left_side_mask[first_passer_b-1]).is_clear())
            {
               if (!w_protected_pp || b_protected_pp)
               {
                  b_outside++;
                  goto step3;
               }
            }
         }
         if (last_passer_b >= 5)
         {
            if (!Bitmap::and(all_pawns,left_side_mask[last_passer_b-1]).is_clear()
                &&  Bitmap::and(board.pawn_bits[White],right_side_mask[last_passer_b-1]).is_clear())
            {
               if (!w_protected_pp || b_protected_pp)
               {
                  b_outside++;
               }
            }
         }
      }
   }
   step3:
   // if there is no outside pawn, then protected pawns are very strong
   if (w_outside && !b_outside)
   {
      score += OUTSIDE_PP[ material_level[Black]];
   } 
   else if (b_outside && !w_outside)
   {
      score -= OUTSIDE_PP[material_level[White]];
   }
   if (w_outside == 0 && b_outside == 0)
   {
      if (w_protected_pp && !b_protected_pp)
      {
         score += OUTSIDE_PP[material_level[Black]];
      }
      if (b_protected_pp && !w_protected_pp)
      {
         score -= OUTSIDE_PP[material_level[White]];
      }
   }
   return score;
}

void Scoring::init(const Board &board)
{
   if (!done_init)
   {
      init_bitmaps();
      done_init++;
   }
   if (!w_pawn_hash_table)
   {
      // fixed size for now
      //w_pawn_hash_table = new Pawn_Entry[PAWN_HASH_SIZE];
      //b_pawn_hash_table = new Pawn_Entry[PAWN_HASH_SIZE];
      w_pawn_hash_table = (Pawn_Entry*)(new char[PAWN_HASH_SIZE*sizeof(Pawn_Entry)]);
      b_pawn_hash_table = (Pawn_Entry*)(new char[PAWN_HASH_SIZE*sizeof(Pawn_Entry)]);
      memset(w_pawn_hash_table,0xff,PAWN_HASH_SIZE*sizeof(Pawn_Entry));
      memset(b_pawn_hash_table,0xff,PAWN_HASH_SIZE*sizeof(Pawn_Entry));
      king_pawn_hash_table = (King_Pawn_Entry*)(new char[KING_PAWN_HASH_SIZE*sizeof(King_Pawn_Entry)]);
      memset(king_pawn_hash_table,0xff,KING_PAWN_HASH_SIZE*sizeof(King_Pawn_Entry));
  }
  side_to_move = board.Side();
}      


void Scoring::cleanup()
{
   if (done_init)
   {
      delete [] w_pawn_hash_table;
      delete [] b_pawn_hash_table;
	  delete [] king_pawn_hash_table;
   }
}

BOOL Scoring::in_endgame( const Board &board )
{
   int w_pcount = board.getMaterial(White).pawn_count();
   int b_pcount = board.getMaterial(Black).pawn_count();
   return BOOL(
              (board.getMaterial(White).value() - w_pcount*64 <= 2640 /* 2 rooks + king */) &&
              (board.getMaterial(Black).value() - b_pcount*64 <= 
               2640 /* 2 rooks + king */));
}

// adjust the material score. "ourmat" is the side with the
// material advantage
static int adjust_mat_score(int mdiff, const Material &ourmat,
                            const Material &oppmat)
{  
   int nonpawns = 
   (int)(ourmat.value() - ourmat.pawn_count()*64);
   if (nonpawns < 4000)
   {
      int max = 80 - ((nonpawns-2000)/25);
      return mdiff + (max*mdiff)/(9*64);
   }
   return mdiff;
}

static int MaterialScore( const Board &board, const ColorType side )
{
   // Computes material score, from the perspective of 'side'.
   int mdiff;

   const Material &ourmat = board.getMaterial(side);
   const Material &oppmat = board.getMaterial(OppositeColor(side));
   mdiff =  (int)(ourmat.value() - oppmat.value());
   // If we are ahead in material give a bonus for trading down
   // pieces (but not pawns).
   // But don't do so if we are so low in material that we can't 
   // mate.
   // Also give a lower bonus for pawnless endings.
   //
   if (ourmat.no_pawns() && oppmat.no_pawns())
   {
      if (oppmat.king_only() && mdiff > 0)
      {
         if (ourmat.value() >= 2320 &&
             ((unsigned)ourmat.infobits() != Material::KNN)) // not KNN vs K
         {
            return mdiff + mdiff/10; // bonus because we are sure to mate
         } else
            return (mdiff*2)/10; // not enough material to mate
      }
      else if (ourmat.king_only() && mdiff < 0)
      {
         if (oppmat.value() >= 2320 &&
             ((unsigned)oppmat.infobits() != Material::KNN)) // not KNN vs K
         {
            return mdiff - mdiff/10; // bonus because we are sure to mate
         } else
            return (mdiff*2)/10; // not enough material to mate
      }
/**      else if (Util::Abs(mdiff) <= 4*Constants::PawnValue)
      {
         // Pawnless endings are hard to win, unless we are way
         // ahead.
         if (mdiff > 0)
            return mdiff - Util::Min(48,mdiff/4);
         else if (mdiff < 0)
            return mdiff - Util::Max(-48,mdiff/4);
         else
            return 0;
      }**/
   }
   if (ourmat.no_pawns() && mdiff > 0 && (ourmat.value()< 2320 ||
        ((unsigned)ourmat.infobits() == Material::KNN)))
   {
       return (mdiff*4)/10; // not enough material to mate
   }
   if (oppmat.no_pawns() && mdiff < 0 && (oppmat.value()< 2320 ||
        ((unsigned)oppmat.infobits() == Material::KNN)))
   {
       return (mdiff*4)/10; // not enough material to mate
   }
   // Penalize bad trades. Algorithm similar to Crafty's.
   int score = 0;
   int oppMinors = oppmat.bishop_count() + oppmat.knight_count();
   int ourMinors = ourmat.bishop_count() + ourmat.knight_count();
   if (ourMinors != oppMinors)
   {
     int ourQ = ourmat.queen_count();
     int oppQ = oppmat.queen_count();
     if (ourQ == oppQ && ourmat.rook_count() == oppmat.rook_count())
     {
       // penalize knight or bishop traded away for pawns
       if (ourMinors > oppMinors) score-=BAD_TRADE;
       else  score+=BAD_TRADE;
     }
     else if (ourQ != oppQ)
     {  
        // penalize Q traded for Rook + Knight or bishop
        if (ourQ == 0)
           score += BAD_TRADE;
        else
           score -= BAD_TRADE;
     }
     else
     {
       // penalize rook traded for knight+bishop
       if (ourMinors > oppMinors+1) score-=BAD_TRADE;
       else if (oppMinors > ourMinors+1) score+=BAD_TRADE;
     }
   }

   // Give a proportional bonus if we are substantially ahead (or
   // a penalty if we are behind) in material (but not if there
   // are zero pawns).
   if (mdiff > 2*(int)Constants::PawnValue)
   {
      return score+adjust_mat_score(mdiff,ourmat,oppmat);
   } 
   else if (mdiff < -2*(int)Constants::PawnValue)
   {
      return score-adjust_mat_score(-mdiff,oppmat,ourmat);
   } 
   else
      return score+mdiff;
}

static int pawn_structure_score(const Board &board, ColorType side, 
                                Pawn_Entry &entr)
{
   // This function computes a pawn structure score that depends
   // only on the location of pawns (of both colors)
   //
   entr.score = 0;
   entr.passers_count = 0;
   entr.protected_pp = 0;
   entr.first_passer = 9;
   entr.last_passer = 0;
   entr.center_score = 0;
   entr.hc = board.pawn_hash();
   entr.weak_files = 0;
   entr.open_files = 0xff;
   entr.w_square_pawns = entr.b_square_pawns = 0;
   int isolated_count = 0;

   int score = 0;
   int incr = (side == White) ? -8 : 8;
   const ColorType oside = OppositeColor(side);
   entr.opponent_pawn_attacks = board.all_pawn_attacks(oside);
   Piece enemy_pawn = MakePiece(Pawn,oside);
   Piece our_pawn = MakePiece(Pawn,side);
   int doubled_pawns = 0;
   Bitmap bi(board.pawn_bits[side]);
   Square sq;
   int lside = 0, rside = 0;
   for (;(sq = bi.first_one()) != InvalidSquare;bi.clear(sq))
   {
#ifdef PAWN_DEBUG
      int tmp = score;
#endif
      int file = File(sq);
      int rank = Rank(sq,side);
      if (file < 4)
      {
        lside++;
      }
      else if (file > 5)
      {
        rside++;
      }
      if (SquareColor(sq) == White)
        entr.w_square_pawns++;
      else
        entr.b_square_pawns++;
      entr.open_files &= (byte)~(1<<(file-1));
      BOOL backward = FALSE;
      BOOL passed = FALSE;
      BOOL weak = FALSE;
      BOOL doubled = FALSE;
      BOOL isolated;
      if (side == White)
      {
         entr.center_score += PawnCenterScores[sq];
      } 
      else
      {
         entr.center_score += PawnCenterScores[63-(int)sq];
      }
      Bitmap result;
      if (side == White)
         result = Bitmap::and(board.PawnBits(side),backwardW[sq]);
      else
         result = Bitmap::and(board.PawnBits(side),backwardB[sq]);
      backward = result.is_clear();
      if (side == White)
         result = Bitmap::and(board.PawnBits(oside),passedW[sq]);
      else
         result = Bitmap::and(board.PawnBits(oside),passedB[sq]);
      passed = result.is_clear();
      if (file == 1)
      {
         isolated = Bitmap::and(
            board.pawn_bits[side],Bearing::file_mask[1]).is_clear();
      }
      else if (file == 8)
      {
         isolated = Bitmap::and(board.pawn_bits[side],
            Bearing::file_mask[6]).is_clear();
      }
      else
      {
         isolated = Bitmap::and(board.pawn_bits[side],
            Bearing::file_mask[file-2]).is_clear() &&
                     Bitmap::and(board.pawn_bits[side],
            Bearing::file_mask[file]).is_clear();
      }
      if (isolated) isolated_count++;
      int pr;
      int fileCount = 1;
      if ((pr = PFileCount(board,file-1,side)) > 1)
      {
         fileCount = pr;
         doubled_pawns++;
         doubled = TRUE;
         if (pr == 2)
            score += DOUBLED_PAWNS/2;
         else
            score += TRIPLED_PAWNS/3;
      }
      if (board[sq+incr] != enemy_pawn)
      {
        int defenders = board.pawn_attacks(sq,side).bit_count();
        if (defenders == 0)
        {
           entr.weak_files |= (1<<(file-1));
           Square sq2(sq + incr);
           int patcks = board.pawn_attacks(sq2,oside).bit_count() -
                        board.pawn_attacks(sq2,side).bit_count();
           if (patcks > 0)
           {
              score += WEAK;
              weak++;
              if (Bitmap::and(board.pawn_bits[oside],
                  Bearing::file_mask[file-1]).is_clear())
                 score += WEAK_ON_OPEN_FILE;
           }
           if (patcks > 1) // "extra weak" pawn
              score += WEAK2;
        }
      }
      if (!isolated && !weak && !backward && !passed && rank > 2)
      {
          //score += UNSUPPORTED_PAWN;
          if (rank >4) score += ADV_UNSUPPORTED_PAWN;
      }
      if (passed)
      {
         entr.passers[entr.passers_count++] = sq;
         ASSERT(entr.passers_count < 8);
         if (file > entr.last_passer)
            entr.last_passer = file;
         if (file < entr.first_passer)
            entr.first_passer = file;
         score += PASSED_PAWN[rank-1];
         if (side == White)
         {
            Bitmap atcks(Bearing::pawn_attacks[sq][White]);
            atcks.and(board.pawn_bits[White]);
            if (!atcks.is_clear())
               entr.protected_pp = 1;
         } 
         else
         {
            Bitmap atcks(Bearing::pawn_attacks[sq][Black]);
            atcks.and(board.pawn_bits[Black]);
            if (!atcks.is_clear())
               entr.protected_pp = 1;
         }
         int j;
         for (j=0;j<entr.passers_count-1;j++)
         {
            Square psq = entr.passers[j];
            int psqfile = File(psq);
            if (psqfile == file-1 || psqfile == file+1)
            {
               int psqrank = Rank(psq,side);
               int rdiff = rank - psqrank;
               // score is based on the more advanced
               // of the 2 pawns
               if (rdiff == 0 || rdiff == 1)
               {
                  score += CONNECTED_PASSERS[rank-1];
               } 
			   else if (rdiff == -1)
               {
                  score += CONNECTED_PASSERS[psqrank-1];
               }
            }
         }
         if (file ==1 || file == 8)
            score += (RIM_PASSED_PAWN*(rank+1))/8;
         for (j=rank+1;j<9;j++)
         {
            Square sq2 = Square(file,j,side);
            Piece piece2 = board[sq2];
            if (TypeOfPiece(piece2) != EmptyPiece)
            {
               if (PieceColor(piece2) == side)
               {
                  if (TypeOfPiece(piece2) == Pawn)
                  {
                     // This pawn is passed but another
                     // pawn is ahead of it, i.e. it is
                     // doubled
                     score -= PASSED_PAWN[rank-1]/3;
                  }
               }
            }
         }
      }
	  int duo = 0;
      if (rank > 3 && file != 8 && board[sq+1] == our_pawn)
      {
         score += PAWN_DUO;
		 duo++;
      }
      if (!passed && rank >= 4)
      {
         int space = SPACE*(rank-3);
         if (duo || !board.pawn_attacks(sq,side).is_clear())
            space += (SPACE*(rank-3))/2;
         if (rank > 5 && board[sq+incr] == enemy_pawn)
          // cramping enemy pawn
            space += (SPACE*(rank-3))/2;
         score += space;
      }
#ifdef PAWN_DEBUG
      cout << Image(side) << " pawn on " << sq << " score = " << score-tmp << endl;
#endif
   }
   score += ISOLATED_PAWN[isolated_count];
   entr.score = score;
   return score;
}

/*static*/ void calc_cover(const Board &board,Square kp,ColorType side,
                       int &cover, int &proximity)
{
   cover = proximity = 0;
   int kpfile = File(kp);
   const Piece our_pawn = MakePiece(Pawn,side);
   const Piece our_bishop = MakePiece(Bishop,side);
   const ColorType oside = OppositeColor(side);
   int kcover = 0, qcover = 0;

//   if (kpfile > 5)
   {
      Square pawnsq = Square(6,2,side);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(6,3,side);
         if (board[pawnsq2] == our_pawn)
         {
            kcover += BP_COVER_MISSING;
         } 
         else
            kcover += BP_COVER_MISSING2;
      }
      if (FileOpen(5,side))
      {
         if (FileOpen(5,oside))
         {
            kcover += 2*KING_NEAR_OPEN_FILE;
         } 
         else
            kcover += KING_NEAR_OPEN_FILE;
      }
      pawnsq = (int)(pawnsq+1);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(7,3,side);
         if (board[pawnsq2] == our_pawn)
         {
            kcover += NP_COVER_MISSING;
            if (board[Square(7,2,side)] == our_bishop)
              kcover += FIANCHETTO;
         } 
	 else
            kcover += NP_COVER_MISSING2;
      }
      if (FileOpen(6,side))
      {
         if (FileOpen(6,oside))
         {
            kcover += 2*KING_NEAR_OPEN_FILE;
         } 
	 else
            kcover += KING_NEAR_OPEN_FILE;
      }
      pawnsq = (int)(pawnsq+1);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(8,3,side);
         if (board[pawnsq2] == our_pawn)
            kcover += RP_COVER_MISSING;
         else
            kcover += RP_COVER_MISSING2;
      }
      if (FileOpen(7,side))
      {
         if (FileOpen(7,oside))
         {
            kcover += 2*KING_NEAR_OPEN_FILE;
         } 
         else
            kcover += KING_NEAR_OPEN_FILE;
      }
   } 
//   else if (kpfile < 4)
   {
      Square pawnsq = Square(1,2,side);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(1,3,side);
         if (board[pawnsq2] == our_pawn)
            qcover += RP_COVER_MISSING;
         else
            qcover += RP_COVER_MISSING2;
      }
      if (FileOpen(0,side))
      {
         if (FileOpen(0,oside))
         {
            qcover += 2*KING_NEAR_OPEN_FILE;
         } 
         else
            qcover += KING_NEAR_OPEN_FILE;
      }
      pawnsq = (int)(pawnsq+1);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(2,3,side);
         if (board[pawnsq2] == our_pawn)
         {
            qcover += NP_COVER_MISSING;
            if (board[Square(2,2,side)] == our_bishop)
              qcover += FIANCHETTO;
         } 
	 else
            qcover += NP_COVER_MISSING2;
      }
      if (FileOpen(1,side))
      {
         if (FileOpen(1,oside))
         {
            qcover += 2*KING_NEAR_OPEN_FILE;
         } 
	 else
            qcover += KING_NEAR_OPEN_FILE;
      }
      pawnsq = (int)(pawnsq+1);
      if (board[pawnsq] != our_pawn)
      {
         Square pawnsq2 = Square(3,3,side);
         if (board[pawnsq2] == our_pawn)
         {
            qcover += BP_COVER_MISSING;
         }
         else
            qcover += BP_COVER_MISSING2;
      }
      if (FileOpen(2,side))
      {
         if (FileOpen(2,oside))
         {
            qcover += 2*KING_NEAR_OPEN_FILE;
         } 
         else
            qcover += KING_NEAR_OPEN_FILE;
      }
   }
   if (kpfile > 5)
   {
      cover = kcover;
   }
   else if (kpfile < 4)
   {
      cover = qcover;
   }
   else //if (kpfile == 4 || kpfile == 5)
   {
      cover += (kcover + qcover)/2;
      if (side == White)
      {
         Bitmap b(center_pawn_cover_w);
         b.and(board.pawn_bits[White]);
         cover += 2*((int)b.bit_count()-4);
      } 
      else
      {
         Bitmap b(center_pawn_cover_b);
         b.and(board.pawn_bits[Black]);
         cover += 2*((int)b.bit_count()-4);
      }
      if (FileOpen(3,side))
      {
         if (FileOpen(3,oside))
         {
            cover += 2*KING_NEAR_OPEN_FILE;
         } 
         else
            cover += KING_NEAR_OPEN_FILE;
      }
      if (FileOpen(4,side))
      {
         if (FileOpen(4,oside))
         {
            cover += 2*KING_NEAR_OPEN_FILE;
         } 
         else
            cover += KING_NEAR_OPEN_FILE;
      }
//      cover *= 2;
   }
   if (side == White)
   {
      Bitmap patcks(w_pawn_attacks[kpfile-1]);
      // check for enemy pawns near the king
      patcks.and(board.pawn_bits[Black]);
      if (!patcks.is_clear())
      {
         Bitmap p1(patcks);
         p1.and(Bearing::rank_mask[6]);
         proximity -= 11*p1.bit_count();
         p1 = patcks;
         p1.and(Bearing::rank_mask[5]);
         proximity -= 8*p1.bit_count();
         p1 = patcks;
         p1.and(Bearing::rank_mask[4]);
         proximity -= 5*p1.bit_count();
         p1 = patcks;
         p1.and(Bearing::rank_mask[3]);
         proximity -= 2*p1.bit_count();
      }
   } 
   else
   {
      // check for enemy pawns near the king
      Bitmap patcks(b_pawn_attacks[kpfile-1]);
      patcks.and(board.pawn_bits[White]);
      if (!patcks.is_clear())
      {
         Bitmap p1(patcks);
         p1.and(Bearing::rank_mask[1]);
         proximity -= 11*p1.bit_count();
         p1 = patcks;
         p1.and(Bearing::rank_mask[2]);
         proximity -= 8*p1.bit_count();
         p1 = patcks;
         p1.and(Bearing::rank_mask[3]);
         proximity -= 5*p1.bit_count();
         p1 = patcks;
         p1.and(Bearing::rank_mask[4]);
         proximity -= 2*p1.bit_count();
      }
   }
}

/*static*/ int battery_score(const Board &board, Bitmap &rq,
                         const ColorType side, const int file)
{
   int score = 0;
   // There are rooks and/or queens on a file near the king.  See
   // if there are pawns in the way.
   const Square kp = board.KingPos(side);   
   Piece ourpawn = MakePiece(Pawn,side);
   const ColorType oside = OppositeColor(side);
   Piece enemypawn = MakePiece(Pawn,oside);
   int stack = 0;
   int mate = 0;
   int intervening = 0;
   Square sq;
   for (;(sq = rq.first_one()) != InvalidSquare;rq.clear(sq))
   {
      Square ksq = Square(file,Rank(kp,side),side);
      Square x = sq;
      int ourPawns = 0, enemyPawns = 0;
      if ((int)ksq > (int)sq)
      {
         while ((int)x < 63 && (int)x < (int)ksq)
         {
            x += 8;
            if (board[x] == ourpawn)
               ourPawns++;
            else if (board[x] == enemypawn)
               enemyPawns++;
            else if (board.rooksqueens[oside].is_set(x))
            {
               if (enemyPawns == 0 && ourPawns == 0)
                  ++stack;
            }
            else if (board[x] != EmptyPiece)
            {
               ++intervening;
            }
         }
      } 
      else
      {
         while ((int)x >= 8 && (int)x > (int)ksq)
         {
            x -= 8;
            if (board[x] == ourpawn)
               ourPawns++;
            else if (board[x] == enemypawn)
               enemyPawns++;
            else if (board.rooksqueens[oside].is_set(x))
            {
               if (enemyPawns == 0 && ourPawns == 0)
                  ++stack;
            }
            else if (board[x] != EmptyPiece)
            {
               ++intervening;
            }
         }
      }
      if (enemyPawns && ourPawns)
      {
         score += BATTERY0;
      } 
      else if (enemyPawns)
         score += BATTERY1;
      else
      {
         if (ourPawns)
            score += BATTERY2;
         else
            score += BATTERY3;
      }
   }
   if ((file == 1 || file == 8) && side_to_move == side)
   {
      if (stack && score <= 2*BATTERY3 && //kp.Rank(side)==1 &&
          !Bitmap::and(board.queen_bits[oside],Bearing::file_mask[file-1]).is_clear())
      {
         // see if the king has a possible escape square
         Square escape;
         if (side == White)
         {
            if (file == 1)
               escape = 50;
            else
               escape = 53;
         } 
         else
         {
            if (file == 1)
               escape = 10;
            else
               escape = 13;
         }
         if (IsEmptyPiece(board[escape]))
            mate += POTENTIAL_ROOK_FILE_MATE-intervening*10;
         else
            mate += ROOK_FILE_MATE_THREAT-intervening*10;
      } 
      else if (score <= BATTERY3 && 
                 !Bitmap::and(board.rook_bits[oside],Bearing::file_mask[file-1]).is_clear() &&
                 !board.queen_bits[oside].is_clear())
         mate += POTENTIAL_ROOK_FILE_MATE-intervening*10;
   }
   static const int stack_factor[6] = { 1, 5, 10, 15, 15, 15};
#ifdef EVAL_DEBUG
   if (side==White)
     cout << "battery (w): score = " << score << " mate= " << mate << 
       " stack = " << stack << endl;
   else
     cout << "battery (b): score = " << score << " mate= " << mate << 
       " stack = " << stack << endl;
#endif
   return (stack == 0) ? score+mate : mate+(stack_factor[stack]*score)/8;
}

#define CHECK_BATTERY(board,file,side,oside) \
{ \
   Bitmap rq(board.rooksqueens[oside]); \
   rq.and(Bearing::file_mask[file-1]); \
   if (!rq.is_clear() && (FileOpen(file-1,side) || \
       rq.bit_count() > 1)) \
      proximity += battery_score(board,rq,side,file); \
}

int Scoring::evalu8( const Board &board )
{
   endgame = in_endgame(board);
   if (endgame && is_draw(board))
      return 0;
   else
   {
#ifdef EVAL_DEBUG
     int mat_score = MaterialScore(board,board.Side());
     cout << "material score = " << mat_score << endl;
     return mat_score +
      positional_score(board,-Constants::BIG,Constants::BIG);
#else
      return MaterialScore(board,board.Side()) +
      positional_score(board,-Constants::BIG,Constants::BIG);
#endif
   }
}

BOOL Scoring::is_draw(const Board &board)
{
   return (repetition_draw(board) || material_draw(board) ||
           is_theoretical_draw(board));
}        

BOOL Scoring::is_legal_draw(const Board &board)
{
   return (repetition_draw(board) || material_draw(board));
}        

BOOL Scoring::is_theoretical_draw(const Board &board)
{
   const Material &mat1 = board.getMaterial(White);
   const Material &mat2 = board.getMaterial(Black);
   if (mat1.value() > 2000+(BISHOP_VALUE*2) &&
       mat2.value() > 2000+(BISHOP_VALUE*2))
       return FALSE;
   // Check for K + P vs rook pawn
   if ((unsigned)mat1.infobits() == Material::KP &&
       mat2.king_only())
   {
      if (!Bitmap::and(rook_pawn_mask,
                       board.pawn_bits[White]).is_clear())
      {
         Square kp = board.KingPos(White);
         Square kp2 = board.KingPos(Black);
         Square psq = (board.pawn_bits[White].first_one());
         if (Rank(kp,White) < Rank(kp2,White))
         {
            Square qsq = Square(File(psq),8,White);
            return !uncatchable(board,psq,kp,kp2,White) &&
            distance(kp2,qsq) < distance(kp,qsq);
         }
      }
   } 
   else if ((unsigned)mat2.infobits() == Material::KP &&
              mat1.king_only())
   {
      if (!Bitmap::and(rook_pawn_mask,
                       board.pawn_bits[Black]).is_clear())
      {
         Square kp = board.KingPos(Black);
         Square kp2 = board.KingPos(White);
         Square psq(board.pawn_bits[Black].first_one());
         if (Rank(kp,Black) < Rank(kp2,Black))
         {
            Square qsq = Square(File(psq),8,Black);
            return !uncatchable(board,psq,kp,kp2,Black) &&
            distance(kp2,qsq) < distance(kp,qsq);
         }
      }
   }
   // Check for wrong bishop + rook pawn vs king.  Not very common but
   // we don't want to get it wrong.
   else if (mat1.infobits() == Material::KBP && mat2.king_only())
   {
      Bitmap pb(board.pawn_bits[White]);
      if (!(Bitmap::and(pb,Bearing::file_mask[0])).is_clear())
      {
         // on the a file
         // queen square is white
         if ((Bitmap::and(board.bishop_bits[White],Bearing::white_squares)).is_clear())
         {
            // wrong color bishop
            Square pawnsq = pb.first_one();
            if (Rank(board.KingPos(Black),White) >= Rank(pawnsq,White))
            {
               if (Util::Abs((int)File(board.KingPos(Black)) - (int)File(pawnsq)) <= 1)
               {
                  return TRUE;
               }
            }
         }
      } 
      else if (!(Bitmap::and(pb,Bearing::file_mask[7])).is_clear())
      {
         // on the h file
         // queen square is black
         if ((Bitmap::and(board.bishop_bits[White],Bearing::black_squares)).is_clear())
         {
             // wrong color bishop
             Square pawnsq = pb.first_one();
             if (Rank(board.KingPos(Black),White) >= Rank(pawnsq,White))
             {
                if (Util::Abs((int)File(board.KingPos(Black)) - (int)File(pawnsq)) <= 1)
                {
                   return TRUE;
                }
             }
         }
      }
   } 
   else if (mat2.infobits() == Material::KBP && mat1.king_only())
   {
      Bitmap pb(board.pawn_bits[Black]);
      if (!(Bitmap::and(pb,Bearing::file_mask[0])).is_clear())
      {
         // on the a file
         // queen square is black
         if ((Bitmap::and(board.bishop_bits[Black],Bearing::black_squares)).is_clear())
         {
            // wrong color bishop
            Square pawnsq = pb.first_one();
            if (Rank(board.KingPos(White),Black) >= Rank(pawnsq,Black))
            {
               if (Util::Abs((int)File(board.KingPos(White)) - (int)File(pawnsq)) <= 1)
               {
                  return TRUE;
               }
            }
         }
      } 
      else if (!(Bitmap::and(pb,Bearing::file_mask[7])).is_clear())
      {
         // on the h file
         // queen square is white
         if ((Bitmap::and(board.bishop_bits[Black],Bearing::white_squares)).is_clear())
         {
            // wrong color bishop
            Square pawnsq = pb.first_one();
            if (Rank(board.KingPos(White),Black) >= Rank(pawnsq,Black))
            {
               if (Util::Abs((int)File(board.KingPos(White)) - (int)File(pawnsq)) <= 1)
               {
                  return TRUE;
               }
            }
         }
      }
   }
   // check for KNN vs K
   if (((unsigned)mat1.infobits() == Material::KNN &&
        mat2.king_only()) ||
       ((unsigned)mat2.infobits() == Material::KNN &&
        mat1.king_only()))
      return TRUE;

   return FALSE;
}

static int score_endgame(const Board &board, ColorType side, Pawn_Entry *hash_entry)
{
   BOOL have_opposition = FALSE;
   const ColorType oside = OppositeColor(side);
   const Square kp = board.KingPos(side);
   const Square oppkp = board.KingPos(oside);
   int score = 0;
   const Material &my_mat = board.getMaterial(side);
   const Material &opp_mat = board.getMaterial(OppositeColor(side));
   if (board.Side() != side)
   {
      int x = opposition[kp][oppkp];
      if (x > 0)
         score += OPPOSITION*x;
   }
   if ((my_mat.infobits() == Material::KNB) &&
       opp_mat.king_only())
   {
      // KBNK endgame, special case.
      const int opprank = Rank(oppkp,White);
      const int oppfile = File(oppkp);
      Bitmap bishops = board.bishop_bits[side];
      Square sq = bishops.first_one();
      // try to force king to a corner where mate is possible.
      Square evalsq;
      if (SquareColor(sq) == White)
      {
         evalsq = Square(9-opprank, oppfile, White);
      } 
      else
      {
         evalsq = oppkp;
      }
      score += KBNKScores[evalsq];
      // keep the kings close
      score += 10-distance(board.KingPos(White),board.KingPos(Black));
      return score;
   } 
   else if ((my_mat.infobits() & 0xff000) != 0 &&
               (opp_mat.king_only() || 
                opp_mat.infobits() == Material::KN ||
                opp_mat.infobits() == Material::KB))
   {
      // bare K vs. R and/or Q, or Q vs 1 minor piece
      // "side" has the material advantage.
      // Encourage moving the opposing king to the edge.
      score += 2*(10-KingCenterScores[board.KingPos(oside)]);
      // Keep the kings close.
      score += 15-3*distance(board.KingPos(White),board.KingPos(Black));
   } 
   else if ((my_mat.infobits() == Material::KP) &&
              (unsigned)opp_mat.king_only())
   {
      // KPK endgame
      Square passer = hash_entry->passers[0];
      if (uncatchable(board,passer,kp,oppkp,side))
      {
         // The opposing king can't catch the pawn before
         // it queens
         score += (UNCATCHABLE_PAWN*Rank(passer,side))/7;
      } else
      {
         // This should be a draw with correct play, so 
         // bring the score nearer zero:
         score -= 50;
         int rank = Rank(passer,side);
         int file = File(passer);
         int file_dist = Util::Abs(File(passer) - File(kp));
         int rank_dist = Rank(kp,side)-rank;
         // strongly encourage putting the king ahead of the
         // pawn
         if (rank_dist > 0 && file_dist <= 1)
            score += KING_AHEAD_OF_PAWN;
         else // king behind pawn
            score -= PASSED_PAWN[rank-1]/2;
         //also encourage staying near pawn
         score += 10 + DISTANCE_FROM_PAWN*distance(kp,passer);
      }
   } 
   else if (opp_mat.no_pieces() && hash_entry->passers_count)
   {
      // our opponent has no pieces and we have at least one
      // passed pawn
      int uncatchables = 0;
      int minfile = 9;
      int maxfile = 0;
      Square minfilesq, maxfilesq;
      for (int i = 0; i < hash_entry->passers_count; i++)
      {
         Square passer = hash_entry->passers[i];
         if (uncatchable(board,passer,kp,oppkp,side))
         {
            ++uncatchables;
            if (side == White)
            {
               if (Rank(passer,White) > white_passer_rank)
               {
                  white_passer_rank = Rank(passer,White);
                  white_passer_race_rank = white_passer_rank;
                  white_passer_sq = white_passer_race_sq = passer;
               }
            } 
            else
            {
               if (Rank(passer,Black) > black_passer_rank)
               {
                  black_passer_rank = Rank(passer,Black);
                  black_passer_race_rank = black_passer_rank;
                  black_passer_sq = black_passer_race_sq = passer;
               }
            }
         }
         if (File(passer) < minfile)
         {
            minfile = File(passer); minfilesq = passer;
         }
         if (File(passer) > maxfile)
         {
            maxfile = File(passer); maxfilesq = passer;
         }
      }
      if (!uncatchables && hash_entry->passers_count > 1)
      {
         int d21 = distance(oppkp,minfilesq);
         int d22 = distance(oppkp,maxfilesq);
         int x1 = d21 + Rank(maxfilesq,side) -1;
         int x2 = d22 + Rank(minfilesq,side) -1;
         int y = 0;
         if (board.Side() != side) y = 1;
         if (x1+y >= 8 && x2+y >= 8)
         {
            uncatchables++;
            if (side == White)
            {
               int rank = Util::Max(Rank(maxfilesq,White),
                                    Rank(minfilesq,White));
               if (rank > white_passer_rank)
               {
                  white_passer_rank = rank;
                  if (rank == Rank(maxfilesq,White))
                     white_passer_sq = maxfilesq;
                  else
                     white_passer_sq = minfilesq;
               }
            } 
            else
            {
               int rank = Util::Max(Rank(maxfilesq,Black),
                                    Rank(minfilesq,Black));
               if (rank > black_passer_rank)
               {
                  black_passer_rank = rank;
                  if (rank == Rank(maxfilesq,Black))
                     black_passer_sq = maxfilesq;
                  else
                     black_passer_sq = minfilesq;
               }
            }
         } 
	 else if (x2+y < 8 && uncatchable(board,
                                          Square(File(minfilesq),
                                          x2+y,
                                          side),
                                          kp,
                                          maxfilesq,side))
         {
            uncatchables++;
            if (side == White)
            {
               int rank = Util::Max(Rank(maxfilesq,White),
                                    Rank(minfilesq,White));
               if (rank > white_passer_rank)
               {
                  if (rank == Rank(maxfilesq,White))
                     white_passer_sq = maxfilesq;
                  else
                     white_passer_sq = minfilesq;
                  white_passer_rank = rank;
               }
            } 
  	    else
            {
               int rank = Util::Max(Rank(maxfilesq,Black),
                                    Rank(minfilesq,Black));
               if (rank > black_passer_rank)
               {
                  if (rank == Rank(maxfilesq,Black))
                     black_passer_sq = maxfilesq;
                  else
                     black_passer_sq = minfilesq;
                  black_passer_rank = rank;
               }
            }
         } 
	 else if (x1+y < 8 && uncatchable(board,
                                          Square(File(maxfilesq),
                                          x1+y,
                                          side),
                                          kp,
                                          minfilesq,side))
         {
            uncatchables++;
            if (side == White)
            {
               int rank = Util::Max(Rank(maxfilesq,White),
                                    Rank(minfilesq,White));
               if (rank > white_passer_rank)
               {
                  if (rank == Rank(maxfilesq,White))
                     white_passer_sq = maxfilesq;
                  else
                     white_passer_sq = minfilesq;
                  white_passer_rank = rank;
               }
            } 
            else
            {
               int rank = Util::Max(Rank(maxfilesq,Black),
                                    Rank(minfilesq,Black));
               if (rank > black_passer_rank)
               {
                  if (rank == Rank(maxfilesq,Black))
                     black_passer_sq = maxfilesq;
                  else
                     black_passer_sq = minfilesq;
                  black_passer_rank = rank;
               }
            }
         }
      }
      if (!uncatchables)
      {
         if (Rank(kp,side) > 4)
            score += ADVANCED_KING;
         Bitmap all_pawns(Bitmap::or(board.pawn_bits[White],board.pawn_bits[Black]));
         if (Bitmap::and(all_pawns,left_side_mask[4]).is_clear() ||
             Bitmap::and(all_pawns,right_side_mask[3]).is_clear())
         {
            // bonus for king near pawns
            Bitmap it(all_pawns);
            Square sq;
            for (;(sq = it.first_one()) != InvalidSquare;it.clear(sq))
            {
               score += (4-distance(kp,sq));
            }
         } else
            score += KingCenterScores[kp];
      }
   } 
   else
   {
      // This code is VERY crude .. it helps (a little) in king and
      // pawn endgames, but is pretty useless otherwise ... Lots more
      // work to do here.
      if (Rank(kp,side) > 4)
         score += ADVANCED_KING;
      Bitmap all_pawns(Bitmap::or(board.pawn_bits[White],board.pawn_bits[Black]));
      if (Bitmap::and(all_pawns,left_side_mask[4]).is_clear() ||
          Bitmap::and(all_pawns,right_side_mask[3]).is_clear())
      {
         // bonus for king near pawns
         Bitmap it(all_pawns);
         Square sq;
         for (;(sq = it.first_one()) != InvalidSquare;it.clear(sq))
         {
            score += (4-distance(kp,sq));
         }
      } else
         score += KingCenterScores[kp];

      // give a bonus for unstoppable passers depending on the
      // material our opponent has.  (Our opponent has
      // at least one piece).
      int xtramat = opp_mat.value() - 2000;
      Square uncatchable_list[8];
      if (xtramat < 2000)
      {
         int uncatchables = 0;
         for (int i = 0; i < hash_entry->passers_count; i++)
         {
            Square pawnpos = hash_entry->passers[i];
            if (uncatchable(board,pawnpos,kp,oppkp,side))
            {
               uncatchable_list[uncatchables++] = pawnpos;
            } else // bonus for king near passer
               score += (4-distance(kp,pawnpos));
         }
         if (uncatchables)
         {
            int incr = UNCATCHABLE_PAWN2[material_level[board.OppositeSide()]];
            score += incr;
            if (uncatchables > 1)
               score += incr/2;
            if (uncatchables >2)
               score += incr*(uncatchables-2)/4;
         }
      }
   }
   return score;
}

static int king_safety(const Board &board,King_Pawn_Entry *hash_entry)
{
   int score = 0;
   int w_proximity = 0, b_proximity = 0, w_proximity2 = 0, b_proximity2 = 0;
   Square kp = board.KingPos(White);
   if (kp == 56 || kp == 57)
   {
      Piece p = board[40];
      if (!IsEmptyPiece(p) && PieceColor(p) != White
          && (TypeOfPiece(p) == Queen || TypeOfPiece(p) ==
              Bishop || TypeOfPiece(p) == Pawn))
      {
         score += KING_TRAPPED_IN_CORNER;

         if (board[49] == EmptyPiece && 
             Bitmap::and(longdiagDa8,board.bishopsqueens[White]).is_clear() &&
             ((TypeOfPiece(p) == Queen && 
                                         !Bitmap::and(longdiagDa1,
                                                      board.bishopsqueens[Black]).is_clear()) ||
             !Bitmap::and(longdiagDa1,
                          board.queen_bits[Black]).is_clear()))
            score += DIAGONAL_MATE_THREAT;
      }
   } 
   else if (kp == 62 || kp == 63)
   {
      Piece p = board[47];
      if (!IsEmptyPiece(p) && PieceColor(p) != White
          && (TypeOfPiece(p) == Queen || TypeOfPiece(p) == 
              Bishop || TypeOfPiece(p) == Pawn))
      {
         score += KING_TRAPPED_IN_CORNER;
         if (board[54] == EmptyPiece && 
             Bitmap::and(longdiagDa8,board.bishopsqueens[White]).is_clear() &&
             ((TypeOfPiece(p) == Queen &&
                                         !Bitmap::and(longdiagDa8,
                                                      board.bishopsqueens[Black]).is_clear()) ||
             !Bitmap::and(longdiagDa8,
                          board.queen_bits[Black]).is_clear()))
            score += DIAGONAL_MATE_THREAT;
      }
   }
   int kprank = Rank(kp,White);
   if (kprank != 1)
   {
      int king_off_back = KING_OFF_BACK_RANK[material_level[Black]];
      if (kprank > 2)
         king_off_back *= 2;
      score += king_off_back;
   }
   if (material_level[Black] > 2)
   {
      if (kprank < 4)
      {
         Bitmap katcks(king_attacks[quadrants[kp]]);
         katcks.or(king_near_proximity[kp]);
         Bitmap result(Bitmap::and(katcks,board.occupied[Black]));
         result.andNot(board.pawn_bits[Black]);
         int attacks;
         if (result.is_clear())
            attacks = 0;
         else
            attacks = result.bit_count();
         w_proximity += KING_PIECE_PROXIMITY[attacks];
         if
             (board[44] == whitePawn &&
              Bitmap::and(b_stonewall,board.pawn_bits[Black]) == b_stonewall &&
              File(kp) > 4)
         {
            w_proximity += STONEWALL;
         }
      }
   }

   kp = board.KingPos(Black);
   if (kp == 0 || kp == 1)
   {
      Piece p = board[16];
      if (!IsEmptyPiece(p) && PieceColor(p) != Black
          && (TypeOfPiece(p) == Queen || TypeOfPiece(p) ==
              Bishop || TypeOfPiece(p) == Pawn))
      {
         score -= KING_TRAPPED_IN_CORNER;
         if (board[9] == EmptyPiece && 
             Bitmap::and(longdiagDa8,board.bishopsqueens[Black]).is_clear() &&
             ((TypeOfPiece(p) == Queen &&
                                        !Bitmap::and(longdiagDa8,
                                                     board.bishopsqueens[White]).is_clear()) ||
             !Bitmap::and(longdiagDa8,
                          board.queen_bits[White]).is_clear()))
            score -= DIAGONAL_MATE_THREAT;
      }
   } 
   else if (kp == 6 || kp == 7)
   {
      Piece p = board[23];
      if (!IsEmptyPiece(p) && PieceColor(p) != Black
          && (TypeOfPiece(p) == Queen || TypeOfPiece(p) == 
              Bishop || TypeOfPiece(p) == Pawn))
      {
         score -= KING_TRAPPED_IN_CORNER;
         if (board[15] == EmptyPiece && 
             Bitmap::and(longdiagDa8,board.bishopsqueens[Black]).is_clear() &&
             ((TypeOfPiece(p) == Queen &&
                                         !Bitmap::and(longdiagDa1,
                                                      board.bishopsqueens[White]).is_clear()) ||
             !Bitmap::and(longdiagDa1,
                          board.queen_bits[White]).is_clear()))
            score -= DIAGONAL_MATE_THREAT;
      }
   }
   kprank = Rank(kp,Black);
   if (kprank != 1)
   {
      int king_off_back = KING_OFF_BACK_RANK[material_level[White]];
      if (kprank > 2)
         king_off_back *= 2;
      score -= king_off_back;
   }
   // Fix for C19 attack Bxh7+ Kxh7 Ng5+ Kg6 Qg4
   if (board.CastleStatus(Black) == Board::CastledKSide)
   {
       if (Bitmap::and(board.pawn_bits[Black],
		   Bearing::file_mask[7]).is_clear() &&
           board[20] == blackPawn && board[28] == whitePawn &&
//		   Rank(kp,Black) > 1 &&
           !Bitmap::and(board.queen_bits[White],
            ghfiles).is_clear())
          score += 200;
   }
   if (material_level[White] > 2)
   {
      if (kprank < 4)
      {
         Bitmap katcks(king_attacks[quadrants[kp]]);
         katcks.or(king_near_proximity[kp]);
         Bitmap result(Bitmap::and(katcks,board.occupied[White]));
         result.andNot(board.pawn_bits[White]);
         int attacks;
         if (result.is_clear())
            attacks = 0;
         else
            attacks = result.bit_count();
         b_proximity += KING_PIECE_PROXIMITY[attacks];
         if
             (board[20] == blackPawn &&
              Bitmap::and(w_stonewall,board.pawn_bits[White]) == w_stonewall &&
              File(kp) > 4)
         {
            b_proximity += STONEWALL;
         }
      }
   }

   Square kpfile;
   int proximity;
   if (material_level[Black] > 2)
   {
      proximity = hash_entry->white_pawn_proximity;
      kpfile = File(board.KingPos(White));
      if (kpfile + 1 < 9)
      {
         CHECK_BATTERY(board,kpfile+1,White,Black);
      }
      CHECK_BATTERY(board,kpfile,White,Black);
      if (kpfile - 1 > 0)
      {
         CHECK_BATTERY(board,kpfile-1,White,Black);
      }
      if (kpfile == 8)
         CHECK_BATTERY(board,6,White,Black);
      if (kpfile == 1)
         CHECK_BATTERY(board,3,White,Black);
      w_proximity2 = proximity;
      int factor  = KS_MATERIAL_SCALE[material_level[Black]];
      int cover = (hash_entry->white_cover*factor)/16;
      if (cover <= -30) proximity = -((w_proximity2+w_proximity)*Util::Max(-50,cover))/30;
      else proximity = w_proximity2+w_proximity;
#ifdef EVAL_DEBUG
      cout << "king safety(w) proximity = " << proximity <<
      " cover = " << hash_entry->white_cover << " mlevel = " <<
      material_level[Black] << endl;
#endif
      score += cover + (proximity*factor)/16;
   }
   if (material_level[White] > 2)
   {
      proximity = hash_entry->black_pawn_proximity;
      kpfile = File(board.KingPos(Black));
      if (kpfile + 1 < 9)
      {
         CHECK_BATTERY(board,kpfile+1,Black,White);
      }
      CHECK_BATTERY(board,kpfile,Black,White);
      if (kpfile - 1 > 0)
      {
         CHECK_BATTERY(board,kpfile-1,Black,White);
      }
      if (kpfile == 8)
         CHECK_BATTERY(board,6,Black,White);
      if (kpfile == 1)
         CHECK_BATTERY(board,3,Black,White);
      b_proximity2 = proximity;
      int factor  = KS_MATERIAL_SCALE[material_level[White]];
      int cover = (hash_entry->black_cover*factor)/16;
      if (cover <= -30) proximity = -((b_proximity2+b_proximity)*Util::Max(-50,cover))/30;
      else proximity = b_proximity2+b_proximity;
#ifdef EVAL_DEBUG
      cout << "king safety(b) proximity = " << proximity <<
      " cover = " << hash_entry->black_cover << " mlevel = " <<
      material_level[White] << endl;
#endif
      score -= cover + (proximity*factor)/16;
   }

/**
#ifdef EVAL_DEBUG
   cout << "king safety before adjustment = " << score << endl;
#endif   
   if (w_proximity + w_proximity2 > -10)
   {
      if (b_proximity + b_proximity2 < -20)
      {
         score -= (20 + b_proximity + b_proximity2)/4;
      }
   }
   if (b_proximity + b_proximity2 > -10)
   {
      if (w_proximity + w_proximity2 < -20)
      {
         score += (20 + w_proximity + w_proximity2)/4;
      }
   }
   **/
#ifdef EVAL_DEBUG
   cout << "net king safety score = " << score << endl;
#endif   
   return score;
}

#define PAWN_RACE_WINNER(rank) (UNCATCHABLE_PAWN)

int Scoring::positional_score( const Board &board, int alpha, int beta )
// returns a positional score. Note: scores are from
// the perspective of the White side.
{
   BOOL endgame = in_endgame(board);
   int score = 0;

   // Calculate material level (0-12 for each side)
   material_level[White] = Util::Min(12,board.getMaterial(White).material_level());
   material_level[Black] = Util::Min(12,board.getMaterial(Black).material_level());

   // Query the pawn hash table to see if we have found a position
   // before that has the same pawn structure.  If so, we can skip
   // some of the scoring calculations.

   hash_entry_w = &w_pawn_hash_table[quickmod(board.pawn_hash(),PAWN_HASH_SIZE)];
   hash_entry_b = &b_pawn_hash_table[quickmod(board.pawn_hash(),PAWN_HASH_SIZE)];
   if (hash_entry_w->hash_code() != board.pawn_hash())
   {
      // pawn position not found, calculate the data we need
      pawn_structure_score(board,White,*hash_entry_w);
      pawn_structure_score(board,Black,*hash_entry_b);
#ifdef EVAL_DEBUG
      cout << "white pawn score = " << hash_entry_w->score << endl;
      cout << "black pawn score = " << hash_entry_b->score << endl;
#endif
   }
   first_passer_w = hash_entry_w->first_passer;
   last_passer_w = hash_entry_w->last_passer;
   first_passer_b = hash_entry_b->first_passer;
   last_passer_b = hash_entry_b->last_passer;
   w_protected_pp = hash_entry_w->protected_pp;
   b_protected_pp = hash_entry_b->protected_pp;
   score += hash_entry_w->score - hash_entry_b->score;   
   if (!endgame)
      score += hash_entry_w->center_score - hash_entry_b->center_score;

#ifdef EVAL_DEBUG
   int tmp = score;
#endif
   if (first_passer_w != 9 || first_passer_b != 9)
      score += eval_outside_pp(board);
#ifdef EVAL_DEBUG
   cout << "outside pawn score = " << score - tmp << endl;
   tmp = score;
#endif
   King_Pawn_Entry *kp_hash_entry;
   if (endgame)
   {
#ifdef EVAL_DEBUG
      int tmp = score;
#endif
      white_passer_rank = black_passer_rank = 
      white_passer_race_rank = black_passer_race_rank = 0;
      white_passer_race_sq = black_passer_race_sq = InvalidSquare;
      white_passer_sq = black_passer_sq = InvalidSquare;
#ifdef EVAL_DEBUG
      int endw = score_endgame(board,White,hash_entry_w);
      int endb = score_endgame(board,Black,hash_entry_b);
      cout << "endgame score(w) = " << endw << endl;
      cout << "endgame score(b) = " << endb << endl;
      score += (endw-endb);
      tmp = score;
#else
      score += score_endgame(board,White,hash_entry_w) -
               score_endgame(board,Black,hash_entry_b);
#endif
      Bitmap pawns(board.pawn_bits[White]);
      pawns.or(board.pawn_bits[Black]);
      if (Bitmap::and(pawns,right_mask).data &&
          Bitmap::and(pawns,left_mask).data) {
 	     const Material &w_mat = board.getMaterial(White);
	     const Material &b_mat = board.getMaterial(Black);
         if (material_level[Black] <= 2 &&
			 w_mat.has_bishop() && !b_mat.has_bishop() && b_mat.has_knight())
		 {
            score += BISHOP_ENDGAME_BONUS;
		 }
         else if (material_level[White] <= 2 &&
			 !w_mat.has_bishop() && b_mat.has_bishop() && w_mat.has_knight())
		 {
            score -= BISHOP_ENDGAME_BONUS;
		 }
      }

      // The following pawn race code is effective only when there is only a single
      // passer per side
      if ((material_level[White] == 0 || material_level[Black] == 0))
      {
      if (black_passer_race_rank == 0 && white_passer_race_rank > 0)
         score += PAWN_RACE_WINNER(white_passer_race_rank);
      else if (white_passer_race_rank == 0 && black_passer_race_rank > 0)
         score -= PAWN_RACE_WINNER(black_passer_race_rank);
/**      else if (black_passer_rank == 0 && white_passer_rank > 0)
         score += PAWN_RACE_WINNER(white_passer_rank);
      else if (white_passer_rank == 0 && black_passer_rank > 0)
         score -= PAWN_RACE_WINNER(black_passer_race_rank); **/
      else if (black_passer_race_rank > 0  && white_passer_race_rank > 0)
      {
         ASSERT(black_passer_race_sq != InvalidSquare);
         ASSERT(white_passer_race_sq != InvalidSquare);
         // evaluate pawn race
         // Decrement rank of passer if the king obstructs it
         if (File(board.KingPos(Black))==File(black_passer_race_sq) &&
             Rank(board.KingPos(Black),Black) > black_passer_race_rank)
            --black_passer_race_rank;
         if (File(board.KingPos(White))==File(white_passer_race_sq) &&
             Rank(board.KingPos(White),White) > white_passer_race_rank)
            --white_passer_race_rank;
         if (black_passer_race_rank >= white_passer_race_rank+2 ||
             (black_passer_race_rank == white_passer_race_rank+1 && board.Side() == Black))
            score -= PAWN_RACE_WINNER(black_passer_race_rank);
         else if (white_passer_race_rank >= black_passer_race_rank+2 ||
                  (white_passer_race_rank == black_passer_race_rank+1 && board.Side() == White))
            score += PAWN_RACE_WINNER(white_passer_race_rank);
         else if (white_passer_race_rank == black_passer_race_rank)
         {
            Square white_queen_square = Square(
                                              File(white_passer_race_sq),8,White);
            Square black_queen_square = Square(
                                              File(black_passer_race_sq),8,Black);
            if (board.Side() == White)
            {
               if (white_passer_race_rank == black_passer_race_rank)
               {
                  // White will queen first
                  // See if White will check the black king or
                  // if the White queen will stop the Black
                  // passer.
                  if (board.Clear(white_queen_square,
                                  board.KingPos(Black)) ||
                      (board.Clear(black_queen_square,
                                   board.KingPos(White)) &&
                       distance(board.KingPos(Black),black_queen_square)>1
                      ))
                  {
                     score += PAWN_RACE_WINNER(white_passer_race_rank);
                  }
                  // See if Black will queen with check
                  else if (board.Clear(black_queen_square,
                                       board.KingPos(White)))
                  {
                     score -= PAWN_RACE_WINNER(black_passer_race_rank);
                  }
               }
            } 
            else
            {
               if (black_passer_race_rank == white_passer_race_rank)
               {
                  // Black will queen first.
                  // See if Black will check the White king or
                  // if the Black queen will stop the White
                  // passer.
                  if (board.Clear(black_queen_square,
                                  board.KingPos(White)) ||
                      (board.Clear(white_queen_square,
                                   board.KingPos(Black)) &&
                       distance(board.KingPos(White),white_queen_square)>1
                      ))
                  {
                     score -= PAWN_RACE_WINNER(black_passer_race_rank);
                  }
                  // See if White will queen with check
                  else if (board.Clear(white_queen_square,
                                       board.KingPos(Black)))
                  {
                     score -= PAWN_RACE_WINNER(white_passer_race_rank);
                  }
               }
            }

         }
      }
      }
#ifdef EVAL_DEBUG
      cout << "pawn race score = " << score - tmp << endl;
#endif
   } 
   else
   {
      hash_t kp_hash = Board_Hash::king_pawn_hash(board);
      kp_hash_entry = &king_pawn_hash_table[quickmod(kp_hash,KING_PAWN_HASH_SIZE)];
      if (kp_hash_entry->hc != kp_hash)
      {
         const ColorType side = board.Side();
         calc_cover(board,board.KingPos(White),White,kp_hash_entry->white_cover,
                    kp_hash_entry->white_pawn_proximity);
         calc_cover(board,board.KingPos(Black),Black,kp_hash_entry->black_cover,
                    kp_hash_entry->black_pawn_proximity);
         kp_hash_entry->hc = kp_hash;
      }
      score += king_safety(board,kp_hash_entry);   
   }

   if (board[8] == whiteBishop)
   {
      if (board[17] == blackPawn &&
          board.any_attacks(17,Black))
      {
         score += BISHOP_TRAPPED;
      }
   }
   if (board[15] == whiteBishop)
   {
      if (board[22] == blackPawn &&
          board.any_attacks(22,Black))
      {
         score += BISHOP_TRAPPED;
      }
   }
   if (board[48] == blackBishop)
   {
      if (board[41] == whitePawn &&
          board.any_attacks(41,White))
      {
         score -= BISHOP_TRAPPED;
      }
   }
   if (board[55] == blackBishop)
   {
      if (board[46] == whitePawn &&
          board.any_attacks(46,White))
      {
         score -= BISHOP_TRAPPED;
      }
   }

   // We have evaluated the big parts of the total score. See
   // if we can exit now
#define LAZY1 200
#ifdef LAZY_DEBUG
   int lazy_score = score;
#endif
#ifdef USE_LAZY_EVAL   
//   if (!endgame)
   {
      if (board.Side() == White)
      {
         if (score + LAZY1 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif            
         }
         if (score - LAZY1 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif            
         }
      } 
      else
      {
         if (-score + LAZY1 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif            
         }
         if (-score - LAZY1 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif            
         }
      }
   }
#endif
   if (!endgame)
   {
      /** castling **/
      switch (board.CastleStatus(White))
      {
      case Board::CanCastleEitherSide:
         score += CAN_CASTLE; 
         break;
      case Board::CanCastleKSide:
         score += CANT_CASTLE_Q_SIDE; 
         break;
      case Board::CanCastleQSide:
         score += CANT_CASTLE_K_SIDE; 
         break;
      case Board::CastledKSide:
         score += KCASTLE; break;
      case Board::CastledQSide:
         score += QCASTLE; break;
      case Board::CantCastleEitherSide:
         score += CANT_CASTLE; break;
      }
      switch (board.CastleStatus(Black))
      {
      case Board::CanCastleEitherSide:
         score -= CAN_CASTLE; 
         break;
      case Board::CanCastleKSide:
         score -= CANT_CASTLE_Q_SIDE; 
         break;
      case Board::CanCastleQSide:
         score -= CANT_CASTLE_K_SIDE; 
         break;
      case Board::CastledKSide:
         score -= KCASTLE; break;
      case Board::CastledQSide:
         score -= QCASTLE; break;
      case Board::CantCastleEitherSide:
         score -= CANT_CASTLE; break;
      }
   }

   // interaction of passed pawns and pieces
   const int mlevel_b = material_level[Black];
   const int mlevel_w = material_level[White];

   Square wk = board.KingPos(White);
   for (int i = 0; i < hash_entry_w->passers_count; i++)
   {
      Square sq = hash_entry_w->passers[i];
      ASSERT(OnBoard(sq));
      int file = File(sq);
      int rank = Rank(sq,White);
      // score passed pawns less if large material.
      // But not if opponent has bare king.
      if (!board.getMaterial(Black).king_only())
         score += PASSER_MATERIAL_BONUS[mlevel_b]-6;
      if (!Bitmap::and(FileAttacksUp(board,sq),board.rooksqueens[White]).is_clear())
         score += ROOK_BEHIND_PP[rank-1];
      Square sq2 = sq-8;
      if (!Bitmap::and(FileAttacksDown(board,sq),board.occupied[Black]).is_clear())
      {
         int hit = PASSED_PAWN_BLOCK[rank-1];
         if (!IsEmptyPiece(board[sq2]) && PieceColor(board[sq2]) == Black)
         {
            score -= (hit + PASSED_PAWN_CANT_ADVANCE);
         } else
            score -= hit;
      }
      if (rank == 6 && (File(wk) == file-1 || File(wk) == file+1) &&
          Rank(wk,White) >= rank)
        score += SUPPORTED_PASSER6;
      else if (rank == 7 && (File(wk) == file-1 || File(wk) == file+1) &&
          Rank(wk,White) >= rank)
        score += SUPPORTED_PASSER7;
   }
   Square bk = board.KingPos(Black);
   for (i = 0; i < hash_entry_b->passers_count; i++)
   {
      Square sq = hash_entry_b->passers[i];
      ASSERT(OnBoard(sq));
      int file = File(sq);
      int rank = Rank(sq,Black);
      // score passed pawns less if large material
      if (!board.getMaterial(White).king_only())
         score -= PASSER_MATERIAL_BONUS[mlevel_w]-6;
      if (!Bitmap::and(FileAttacksDown(board,sq),board.rooksqueens[Black]).is_clear())
         score -= ROOK_BEHIND_PP[rank-1];
      Square sq2 = sq+8;
      if (!Bitmap::and(FileAttacksUp(board,sq),board.occupied[White]).is_clear())
      {
         int hit = PASSED_PAWN_BLOCK[rank-1];
         if (!IsEmptyPiece(board[sq2]) && PieceColor(board[sq2]) == White)
            score += (hit + PASSED_PAWN_CANT_ADVANCE);
         else
            score += hit;
      }
      if (rank == 6 && (File(bk) == file-1 || File(bk) == file+1) &&
          Rank(bk,White) >= rank)
        score -= SUPPORTED_PASSER6;
      else if (rank == 7 && (File(bk) == file-1 || File(bk) == file+1) &&
          Rank(bk,White) >= rank)
        score -= SUPPORTED_PASSER7;
   }

#ifdef USE_LAZY_EVAL   
   #define LAZY2 90
#ifdef LAZY_DEBUG
   int lazy_score2=score;
#endif
//   if (!endgame)
   {
      if (board.Side() == White)
      {
         if (score + LAZY2 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif            
         }
         if (score - LAZY2 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif
         }
      } 
      else
      {
         if (-score + LAZY2 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif
         }
         if (-score - LAZY2 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif
         }
      }
   }
#endif

   Bitmap knights(board.knight_bits[White]);
   Square sq;
   for (;(sq = knights.first_one()) != InvalidSquare;knights.clear(sq))
   {
      score += KnightCenterScores[sq];
      if (Rank(sq,White)==1)
         score += KNIGHT_BACK;
      {
         // Compute knight mobility.  We award a bonus for
         // each square to which the knight can move, that
         // is not occupied or attacked by an enemy pawn.
         Bitmap ka(Bearing::knight_attacks[sq]);
         ka.andNot(board.occupied[Black]);
         ka.andNot(board.occupied[White]);
         ka.andNot(hash_entry_w->opponent_pawn_attacks);
         if (!ka.is_clear())
            score += KNIGHT_MOBILITY[ka.bit_count()];
      }
      // evaluate a knight outpost
      if (OutpostScoresW[sq] &&
          Bitmap::and(outpostW[sq],board.pawn_bits[Black]).is_clear())
      {
          score += OutpostScoresW[sq]+KNIGHT_OUTPOST;
          if (!Bitmap::and(Bearing::pawn_attacks[sq][White],
              board.pawn_bits[White]).is_clear())
          {
             score += OutpostScoresW[sq]+KNIGHT_OUTPOST; // defended by pawns
          }
      }
   }
   knights = board.knight_bits[Black];
   for (;(sq = knights.first_one()) != InvalidSquare;knights.clear(sq))
   {
      score -= KnightCenterScores[sq];
      if (Rank(sq,Black)==1)
         score -= KNIGHT_BACK;
      {
         // Compute knight mobility.  We award a bonus for
         // each square to which the knight can move, that
         // is not occupied or attacked by an enemy pawn.
         Bitmap ka(Bearing::knight_attacks[sq]);
         ka.andNot(board.occupied[Black]);
         ka.andNot(board.occupied[White]);
         ka.andNot(hash_entry_b->opponent_pawn_attacks);
         if (!ka.is_clear())
            score -= KNIGHT_MOBILITY[ka.bit_count()];
      }
      // evaluate a knight outpost
      if (OutpostScoresB[sq] &&
          Bitmap::and(outpostB[sq],board.pawn_bits[White]).is_clear())
      {
          score -= OutpostScoresB[sq]+KNIGHT_OUTPOST;
          if (!Bitmap::and(Bearing::pawn_attacks[sq][Black],
              board.pawn_bits[Black]).is_clear())
          {
             score -= OutpostScoresB[sq]+KNIGHT_OUTPOST; // defended by pawns
          }
      }
   }
   Bitmap queens(board.queen_bits[White]);
   if (!queens.is_clear())
   {
      Square sq = queens.first_one();

      if (FileOpen(File(sq)-1,White) &&
          FileOpen(File(sq)-1,Black))
         score += QUEEN_ON_OPEN_FILE;
      if (Rank(sq,White) == 7 && (Rank(board.KingPos(Black),Black)==1 ||
                                  !Bitmap::and(board.pawn_bits[Black],Bearing::rank_mask[1]).is_clear()))
      {
         score += QUEEN_ON_7TH_RANK;
      }
      if (Rank(sq,White) > 2 && material_level[White] > 6)
      {
         Bitmap b(board.knight_bits[White]);
         b.or(board.bishop_bits[White]);
         b.and(Bearing::rank_mask[7]);
         int piecesBack = b.bit_count();
         if (piecesBack > 0)
         {
            if (board.castled(White))
               score += (QUEEN_OUT*(piecesBack+material_level[White]/2))/10;
            else
               score += 3*(QUEEN_OUT*(piecesBack+material_level[White]/2))/20;
         }
      }
      score += QueenCenterScores[(int)sq];
      score += KING_QUEEN_DISTANCE[distance(sq,board.KingPos(Black))];
   }
   queens = board.queen_bits[Black];
   if (!queens.is_clear())
   {
      Square sq = queens.first_one();
      if (FileOpen(File(sq)-1,White) &&
          FileOpen(File(sq)-1,Black))
         score -= QUEEN_ON_OPEN_FILE;

      if (Rank(sq,Black) == 7 && (Rank(board.KingPos(White),White)==1 ||
                                  !Bitmap::and(board.pawn_bits[White],Bearing::rank_mask[6]).is_clear()))
      {
         score -= QUEEN_ON_7TH_RANK;
      }
      if (Rank(sq,Black) > 2 && material_level[Black] > 6)
      {
         Bitmap b(board.knight_bits[Black]);
         b.or(board.bishop_bits[Black]);
         b.and(Bearing::rank_mask[0]);
         int piecesBack = b.bit_count();
         if (piecesBack > 0)
         {
            if (board.castled(Black))
               score -= (QUEEN_OUT*(piecesBack+material_level[Black]/2))/10;
            else
               score -= 3*(QUEEN_OUT*(piecesBack+material_level[Black]/2))/20;
         }
      }
      score -= KING_QUEEN_DISTANCE[distance(sq,board.KingPos(White))];
      score -= QueenCenterScores[(int)sq];
   }

   Bitmap rooks(board.rook_bits[White]);
   for (;(sq = rooks.first_one()) != InvalidSquare;rooks.clear(sq))
   {
      int rank = Rank(sq,White);
      const int file = File(sq);
      if (rank == 7 && (Rank(board.KingPos(Black),Black)==1 ||
                        !Bitmap::and(board.pawn_bits[Black],Bearing::rank_mask[1]).is_clear()))
      {
         score += ROOK_ON_7TH_RANK;
         // find the next piece on the rank
         Bitmap atcks(FileAttacksDown(board,sq));
         atcks.and(board.rooksqueens[White]);
         if (!atcks.is_clear())
            score += TWO_ROOKS_ON_7TH_RANK;
      }
      if (rank == 1 && board.KingPos(White) > 60 &&
          white_rook_trapped.is_set(sq) &&
          !FileOpen(file-1,White))
         score += ROOK_TRAPPED;
      int bonus = 0;
      if (FileOpen(file-1,White) &&
          (FileOpen(file-1,Black) || 
           (hash_entry_b->weak_files & (1<<(file-1)))))
      {
         score += ROOK_ON_OPEN_FILE;

         if (!Bitmap::and(board.rook_bits[White],Bearing::file_mask[file-1]).is_clear())
         {
            if (!Bitmap::and(FileAttacksUp(board,sq),board.rooksqueens[White]).is_clear())
               score += DOUBLED_ROOKS;
         }
      }
   }

   rooks = board.rook_bits[Black];
   for (;(sq = rooks.first_one()) != InvalidSquare;rooks.clear(sq))
   {
      const int rank = Rank(sq,Black);
      const int file = File(sq);
      if (rank == 7 && (Rank(board.KingPos(White),White)==1 ||
                        !Bitmap::and(board.pawn_bits[White],Bearing::rank_mask[6]).is_clear()))
      {
         score -= ROOK_ON_7TH_RANK;
         Bitmap atcks(FileAttacksUp(board,sq));
         atcks.and(board.rooksqueens[Black]);
         if (!atcks.is_clear())
            score -= TWO_ROOKS_ON_7TH_RANK;
      }
      if (rank == 1 && board.KingPos(Black) > 4 &&
          board.KingPos(Black) < 8 &&
          black_rook_trapped.is_set(sq) &&
          !FileOpen(file-1,Black))
         score -= ROOK_TRAPPED;
      int bonus = 0;
      if (FileOpen(file-1,Black) &&
          (FileOpen(file-1,White) || 
           (hash_entry_w->weak_files & (1<<(file-1)))))
      {
         score -= ROOK_ON_OPEN_FILE;

         if (!Bitmap::and(board.rook_bits[Black],Bearing::file_mask[file-1]).is_clear())
         {
            if (!Bitmap::and(FileAttacksDown(board,sq),board.rooksqueens[Black]).is_clear())
               score -= DOUBLED_ROOKS;
         }
      }
   }

#ifdef USE_LAZY_EVAL   
   #define LAZY3 55
#ifdef LAZY_DEBUG
   int lazy_score3=score;
#endif
//   if (!endgame)
   {
      if (board.Side() == White)
      {
         if (score + LAZY3 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif            
         }
         if (score - LAZY3 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return score;
#endif
         }
      } 
      else
      {
         if (-score + LAZY3 < alpha)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " alpha=" << alpha
            << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif
         }
         if (-score - LAZY3 > beta)
         {
#ifdef _TRACE
            cout << "lazy cutoff, score =" << score << " beta=" << beta
            << endl;
            if (Util::Abs(score) > 250)
               cout << board << endl;
#endif
#ifndef LAZY_DEBUG
            return -score;
#endif
         }
      }
   }
#endif


   BOOL haveWbishop = FALSE;
   BOOL haveBbishop = FALSE;
   if (!board.bishop_bits[White].is_clear())
   {
      Bitmap bishops(board.bishop_bits[White]);
      for (;(sq = bishops.first_one()) != InvalidSquare;bishops.clear(sq))
      {
         if (SquareColor(sq) == White)
           haveWbishop = TRUE;
         else
           haveBbishop = TRUE;
         if (OutpostScoresW[sq] &&
              Bitmap::and(outpostW[sq],board.pawn_bits[Black]).is_clear())
         {
            score += OutpostScoresW[sq];
            if (!Bitmap::and(Bearing::pawn_attacks[sq][White],
                board.pawn_bits[White]).is_clear())
            {
               score += OutpostScoresW[sq]; // defended by pawns
            }
         }
         if (!endgame)
         {
            score += BishopPositionW[sq];
         }
         score += BishopCenterScores[sq];
         int mobility = 0;
         int i = sq; int j = 0;
         while (File(i) != 8 && j < 2)
         {
            i -= 7;
            if (i < 0 ) break;
            if (board[i] != EmptyPiece)
            {
               break;
            }
            ++j;
         }
         mobility += j;
         i = sq; j = 0;
         while (File(i) != 1 && j < 2)
         {
            i -= 9;
            if (i < 0) break;
            if (board[i] != EmptyPiece)
            {
               break;
            }
            ++j;
         }
         mobility += j;
         if (File(sq) == 1 && sq+9<64 && board[sq+9] != EmptyPiece)
            mobility -= 1;
         if (File(sq) == 8 && sq+7<64 && board[sq+7] != EmptyPiece)
            mobility -= 1;
         score += BISHOP_MOBILITY[mobility+2];
      } 
   }
   if (haveWbishop && haveBbishop)
      score += TWO_BISHOPS*(16-material_level[Black])/16;
   else if (haveWbishop && !haveBbishop)
   {
      int whitePawns = hash_entry_w->w_square_pawns;
      int blackPawns = hash_entry_w->b_square_pawns;
      if ((whitePawns + blackPawns > 4) && (whitePawns > blackPawns))
         score += BAD_BISHOP;
   } 
   else if (haveBbishop && !haveWbishop)
   {
      int whitePawns = hash_entry_w->w_square_pawns;
      int blackPawns = hash_entry_w->b_square_pawns;
      if ((whitePawns + blackPawns > 4) && (blackPawns > whitePawns))
         score += BAD_BISHOP;
   }
   haveWbishop = haveBbishop = FALSE;

   if (!board.bishop_bits[Black].is_clear())
   {
      Bitmap bishops(board.bishop_bits[Black]);
      Square sq;
      for (;(sq = bishops.first_one()) != InvalidSquare;bishops.clear(sq))
      {
         if (SquareColor(sq) == White)
           haveWbishop = TRUE;
         else
           haveBbishop = TRUE;
         if (OutpostScoresB[sq] &&
              Bitmap::and(outpostB[sq],board.pawn_bits[White]).is_clear())
         {
            score -= OutpostScoresB[sq];
            if (!Bitmap::and(Bearing::pawn_attacks[sq][Black],
                board.pawn_bits[Black]).is_clear())
            {
               score -= OutpostScoresB[sq]; // defended by pawns
            }
         }
         if (!endgame)
         {
            score -= BishopPositionB[sq];
         }
         score -= BishopCenterScores[sq];
         int mobility = 0;
         int i = sq; int j = 0;
         while (File(i) != 8 && j < 2)
         {
            i += 9;
            if (i > 63) break;
            if (board[i] != EmptyPiece)
            {
               break;
            }
            ++j;
         }
         mobility += j;
         i = sq; j = 0;
         while (File(i) != 1 && j < 2)
         {
            i += 7;
            if (i>63) break;
            if (board[i] != EmptyPiece)
            {
               break;
            }
            ++j;
         }
         mobility += j;
         if (File(sq) == 1 && (int)sq-7>0 && board[sq-7] != EmptyPiece)
            mobility -= 1;
         if (File(sq) == 8 && (int)sq-9>0 && board[sq-9] != EmptyPiece)
            mobility -= 1;
         score -= BISHOP_MOBILITY[mobility+2];
      } 
   }

   if (haveWbishop && haveBbishop)
      score -= TWO_BISHOPS*(16-material_level[White])/16;
   if (haveWbishop && !haveBbishop)
   {
      int whitePawns = hash_entry_b->w_square_pawns;
      int blackPawns = hash_entry_b->b_square_pawns;
      if ((whitePawns + blackPawns > 4) && (whitePawns > blackPawns))
         score -= BAD_BISHOP;
   } 
   else if (haveBbishop && !haveWbishop)
   {
      int whitePawns = hash_entry_b->w_square_pawns;
      int blackPawns = hash_entry_b->b_square_pawns;
      if ((whitePawns + blackPawns > 4) && (blackPawns > whitePawns))
         score -= BAD_BISHOP;
   }
   if (!endgame)
   {
      // Don't block center pawns with pieces
      if (board[51] == whitePawn)
      {
         Piece p2 = board[51-8];
         if (TypeOfPiece(p2) != EmptyPiece && 
             TypeOfPiece(p2) != Pawn && 
             PieceColor(p2) == White)
            score += CENTER_PAWN_BLOCK;
      }
      if (board[52] == whitePawn)
      {
         Piece p2 = board[52-8];
         if (TypeOfPiece(p2) != EmptyPiece && 
             TypeOfPiece(p2) != Pawn &&
             PieceColor(p2) == White)
            score += CENTER_PAWN_BLOCK;
      }
      if (board[11] == blackPawn)
      {
         Piece p2 = board[11+8];
         if (TypeOfPiece(p2) != EmptyPiece && 
             TypeOfPiece(p2) != Pawn &&
             PieceColor(p2) == Black)
            score -= CENTER_PAWN_BLOCK;
      }
      if (board[12] == blackPawn)
      {
         Piece p2 = board[12+8];
         if (TypeOfPiece(p2) != EmptyPiece && 
             TypeOfPiece(p2) != Pawn &&
             PieceColor(p2) == Black)
            score -= CENTER_PAWN_BLOCK;
      }
   }

#ifdef LAZY_DEBUG
   if (Util::Abs(lazy_score -score) > LAZY1)
      cout << board << 
       " difference1 = " << Util::Abs(lazy_score -score) << endl;
   if (Util::Abs(lazy_score2 -score) > LAZY2)
      cout << board << 
       " difference2 = " << Util::Abs(lazy_score2 -score) << endl;
   if (Util::Abs(lazy_score3 -score) > LAZY3)
      cout << board << 
       " difference3 = " << Util::Abs(lazy_score3 -score) << endl;
#endif
   if (board.Side() == White)
      return score;
   else
      return -score;
}


int Scoring::material_score( const Board &board )
// returns a material score
{
   return MaterialScore(board,board.Side());
}

int Scoring::material_score( const Board &board, Move m )
{
   Material & oppmat = (Material&)board.getMaterial(board.OppositeSide());
   Material & ourmat = (Material&)board.getMaterial(board.Side());
   // Adjust for trade bonus
   oppmat.remove_piece(Capture(m));
   if (TypeOfMove(m) == Promotion)
   {
     ourmat.remove_piece(Pawn);
     ourmat.add_piece(PromoteTo(m));
   }
   int mat_score = MaterialScore(board,board.Side());
   oppmat.add_piece(Capture(m));
   if (TypeOfMove(m) == Promotion)
   {
     ourmat.add_piece(Pawn);
     ourmat.remove_piece(PromoteTo(m));
   }
   return mat_score;
}

BOOL Scoring::repetition_draw( const Board &board )
{
   return board.rep_count() >= 2;
}

BOOL Scoring::material_draw( const Board &board)
{
   // check for insufficient material
   const Material &mat1 = board.getMaterial(White);
   const Material &mat2 = board.getMaterial(Black);
   if (mat1.pawn_count() || mat2.pawn_count())
   {
      return FALSE;
   }
   if ((mat1.value() <= KING_VALUE + BISHOP_VALUE) &&
       (mat2.value() <= KING_VALUE + BISHOP_VALUE))
   {
      if (mat1.king_only() || mat2.king_only())
         // K vs K, or K vs KN, or K vs KB
         return TRUE;
      else if (mat1.knight_count() || mat2.knight_count())
      {
         // KN vs KN 
         return FALSE;
      } else
      {
         Bitmap wbishops(board.bishop_bits[White]);
         Bitmap bbishops(board.bishop_bits[Black]);
         // drawn if same-color bishops
         if
             (!(Bitmap::and(wbishops,Bearing::black_squares)).is_clear())
         {
            return !(Bitmap::and(bbishops,Bearing::black_squares)).is_clear();
         } else
         {
            return !(Bitmap::and(bbishops,Bearing::white_squares)).is_clear();
         }
      }
   }
   return FALSE;
}

